"""Module for compression methods.

Especially useful for field quantities with high dimensions.

Includes:

- `Compression` — an interface for specifying a compression method for field quantities.
- `SVD` — a Singular Value Decomposition (SVD) compression method.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np
from scipy.interpolate import RBFInterpolator

from amisc.serialize import PickleSerializable

__all__ = ["Compression", "SVD"]


@dataclass
class Compression(PickleSerializable, ABC):
    """Base class for compression methods. Compression methods should:

    - `compute_map` - compute the compression map from provided data
    - `compress` - compress data into a latent space
    - `reconstruct` - reconstruct the compressed data back into the full space
    - `latent_size` - return the size of the latent space
    - `estimate_latent_ranges` - estimate the range of the latent space coefficients

    !!! Note "Specifying fields"
        The `fields` attribute is a list of strings that specify the field quantities to compress. For example, for
        3D velocity data, the fields might be `['ux', 'uy', 'uz']`. The length of the
        `fields` attribute is used to determine the number of quantities of interest at each grid point in `coords`.
        Note that interpolation to/from the compression grid will assume a shape of `(num_pts, num_qoi)` for the
        states on the grid, where `num_qoi` is the length of `fields` and `num_pts` is the length of `coords`. When
        constructing the compression map, this important fact should be considered when passing data to
        `compute_map`.

    In order to use a `Compression` object, you must first call `compute_map` to compute the compression map, which
    should set the private value `self._map_computed=True`. The `coords` of the compression grid must also be
    specified. The `coords` should have the shape `(num_pts, dim)` where `num_pts` is the number of points in the
    compression grid and `dim` is the number of spatial dimensions. If `coords` is a 1d array, then the `dim` is
    assumed to be 1.

    :ivar fields: list of field quantities to compress
    :ivar method: the compression method to use (only svd is supported for now)
    :ivar coords: the coordinates of the compression grid
    :ivar interpolate_method: the interpolation method to use to interpolate to/from the compression grid
                              (only `rbf` (i.e. radial basis function) is supported for now)
    :ivar interpolate_opts: additional options to pass to the interpolation method
    :ivar _map_computed: whether the compression map has been computed
    """
    fields: list[str] = field(default_factory=list)
    method: str = 'svd'
    coords: np.ndarray = None  # (num_pts, dim)
    interpolate_method: str = 'rbf'
    interpolate_opts: dict = field(default_factory=dict)
    _map_computed: bool = False

    @property
    def map_exists(self):
        """All compression methods should have `coords` when their map has been constructed."""
        return self.coords is not None and self._map_computed

    @property
    def dim(self):
        """Number of physical grid coordinates for the field quantity, (i.e. x,y,z spatial dims)"""
        return self.coords.shape[1] if (self.coords is not None and len(self.coords.shape) > 1) else 1

    @property
    def num_pts(self):
        """Number of physical points in the compression grid."""
        return self.coords.shape[0] if self.coords is not None else None

    @property
    def num_qoi(self):
        """Number of quantities of interest at each grid point, (i.e. `ux, uy, uz` for 3d velocity data)."""
        return len(self.fields) if self.fields is not None else 1

    @property
    def dof(self):
        """Total degrees of freedom in the compression grid (i.e. `num_pts * num_qoi`)."""
        return self.num_pts * self.num_qoi if self.num_pts is not None else None

    def _correct_coords(self, coords):
        """Correct the coordinates to be in the correct shape for compression."""
        coords = np.atleast_1d(coords)
        if len(coords.shape) == 1:
            coords = coords[..., np.newaxis] if self.dim == 1 else coords[np.newaxis, ...]
        return coords

    def interpolator(self):
        """The interpolator to use during compression and reconstruction. Interpolator expects to be used as:

        ```python
        xg = np.ndarray    # (num_pts, dim)  grid coordinates
        yg = np.ndarray    # (num_pts, ...)  scalar values on grid
        xp = np.ndarray    # (Q, dim)        evaluation points

        interp = interpolate_method(xg, yg, **interpolate_opts)

        yp = interp(xp)    # (Q, ...)        interpolated values
        ```
        """
        method = self.interpolate_method or 'rbf'
        match method.lower():
            case 'rbf':
                return RBFInterpolator
            case other:
                raise NotImplementedError(f"Interpolation method '{other}' is not implemented.")

    def interpolate_from_grid(self, states: np.ndarray, new_coords: np.ndarray):
        """Interpolate the states on the compression grid to new coordinates.

        :param states: `(*loop_shape, dof)` - the states on the compression grid
        :param new_coords: `(*coord_shape, dim)` - the new coordinates to interpolate to
        :return: `dict` of `(*loop_shape, *coord_shape)` for each qoi - the interpolated states
        """
        new_coords = self._correct_coords(new_coords)
        grid_coords = self._correct_coords(self.coords)
        skip_interp = (new_coords.shape == grid_coords.shape and np.allclose(new_coords, grid_coords))

        ret_dict = {}
        loop_shape = states.shape[:-1]
        coords_shape = new_coords.shape[:-1]
        states = states.reshape((*loop_shape, self.num_pts, self.num_qoi))
        new_coords = new_coords.reshape((-1, self.dim))
        for i, qoi in enumerate(self.fields):
            if skip_interp:
                ret_dict[qoi] = states[..., i]
            else:
                reshaped_states = states[..., i].reshape(-1, self.num_pts).T  # (num_pts, ...)
                interp = self.interpolator()(grid_coords, reshaped_states, **self.interpolate_opts)
                yp = interp(new_coords)
                ret_dict[qoi] = yp.T.reshape(*loop_shape, *coords_shape)

        return ret_dict

    def interpolate_to_grid(self, field_coords: np.ndarray, field_values):
        """Interpolate the field values at given coordinates to the compression grid.

        :param field_coords: `(*coord_shape, dim)` - the coordinates of the field values
        :param field_values: `dict` of `(*loop_shape, *coord_shape)` for each qoi - the field values at the coordinates
        :return: `(*loop_shape, dof)` - the interpolated values on the compression grid
        """
        field_coords = self._correct_coords(field_coords)
        grid_coords = self._correct_coords(self.coords)
        skip_interp = (field_coords.shape == grid_coords.shape and np.allclose(field_coords, grid_coords))

        coords_shape = field_coords.shape[:-1]
        loop_shape = next(iter(field_values.values())).shape[:-len(coords_shape)]
        states = np.empty((*loop_shape, self.num_pts, self.num_qoi))
        field_coords = field_coords.reshape(-1, self.dim)
        for i, qoi in enumerate(self.fields):
            field_vals = field_values[qoi].reshape((*loop_shape, -1))  # (..., Q)
            if skip_interp:
                states[..., i] = field_vals
            else:
                field_vals = field_vals.reshape((-1, field_vals.shape[-1])).T  # (Q, ...)
                interp = self.interpolator()(field_coords, field_vals, **self.interpolate_opts)
                yg = interp(grid_coords)
                states[..., i] = yg.T.reshape(*loop_shape, self.num_pts)

        return states.reshape((*loop_shape, self.dof))

    @abstractmethod
    def compute_map(self, **kwargs):
        """Compute and store the compression map. Must set the value of `coords` and `_is_computed`. Should
        use the same normalization as the parent `Variable` object.

        !!! Note
            You should pass any required data to `compute_map` with the assumption that the data will be used in the
            shape `(num_pts, num_qoi)` where `num_qoi` is the length of `fields` and `num_pts` is the length of
            `coords`. This is the shape that the compression map should be constructed in.
        """
        raise NotImplementedError

    @abstractmethod
    def compress(self, data: np.ndarray) -> np.ndarray:
        """Compress the data into a latent space.

        :param data: `(..., dof)` - the data to compress from full size of `dof`
        :return: `(..., rank)` - the compressed latent space data with size `rank`
        """
        raise NotImplementedError

    @abstractmethod
    def reconstruct(self, compressed: np.ndarray) -> np.ndarray:
        """Reconstruct the compressed data back into the full `dof` space.

        :param compressed: `(..., rank)` - the compressed data to reconstruct
        :return: `(..., dof)` - the reconstructed data with full `dof`
        """
        raise NotImplementedError

    @abstractmethod
    def latent_size(self) -> int:
        """Return the size of the latent space."""
        raise NotImplementedError

    @abstractmethod
    def estimate_latent_ranges(self) -> list[tuple[float, float]]:
        """Estimate the range of the latent space coefficients."""
        raise NotImplementedError

    @classmethod
    def from_dict(cls, spec: dict) -> Compression:
        """Construct a `Compression` object from a spec dictionary."""
        method = spec.pop('method', 'svd').lower()
        match method:
            case 'svd':
                return SVD(**spec)
            case other:
                raise NotImplementedError(f"Compression method '{other}' is not implemented.")


@dataclass
class SVD(Compression):
    """A Singular Value Decomposition (SVD) compression method. The SVD will be computed on initialization if the
    `data_matrix` is provided.

    :ivar data_matrix: `(dof, num_samples)` - the data matrix
    :ivar projection_matrix: `(dof, rank)` - the projection matrix
    :ivar rank: the rank of the SVD decomposition
    :ivar energy_tol: the energy tolerance of the SVD decomposition
    """
    data_matrix: np.ndarray = None          # (dof, num_samples)
    projection_matrix: np.ndarray = None    # (dof, rank)
    rank: int = None
    energy_tol: float = None

    def __post_init__(self):
        """Compute the SVD if the data matrix is provided."""
        if (data_matrix := self.data_matrix) is not None:
            self.compute_map(data_matrix, rank=self.rank, energy_tol=self.energy_tol)

    def compute_map(self, data_matrix: np.ndarray | dict, rank: int = None, energy_tol: float = None):
        """Compute the SVD compression map from the data matrix. Recall that `dof` is the total number of degrees of
        freedom, equal to the number of grid points `num_pts` times the number of quantities of interest `num_qoi`
        at each grid point.

        :param data_matrix: `(dof, num_samples)` - the data matrix. If passed in as a `dict`, then the data matrix
                            will be formed by concatenating the values of the `dict` along the last axis in the order
                            of the `fields` attribute and flattening the last two axes. This is useful for passing
                            in a dictionary of field values like `{field1: (num_samples, num_pts), field2: ...}`
                            which ensures consistency of shape with the compression `coords`.
        :param rank: the rank of the SVD decomposition
        :param energy_tol: the energy tolerance of the SVD decomposition
        """
        if isinstance(data_matrix, dict):
            data_matrix = np.concatenate([data_matrix[field][..., np.newaxis] for field in self.fields], axis=-1)
            data_matrix = data_matrix.reshape(*data_matrix.shape[:-2], -1).T  # (dof, num_samples)

        nan_idx = np.any(np.isnan(data_matrix), axis=0)
        data_matrix = data_matrix[:, ~nan_idx]
        u, s, vt = np.linalg.svd(data_matrix)
        energy_frac = np.cumsum(s ** 2 / np.sum(s ** 2))
        if rank := (rank or self.rank):
            energy_tol = energy_frac[rank - 1]
        else:
            energy_tol = energy_tol or self.energy_tol or 0.95
            idx = int(np.where(energy_frac >= energy_tol)[0][0])
            rank = idx + 1

        self.data_matrix = data_matrix
        self.rank = rank
        self.energy_tol = energy_tol
        self.projection_matrix = u[:, :rank]  # (dof, rank)
        self._map_computed = True

    def compress(self, data):
        return np.squeeze(self.projection_matrix.T @ data[..., np.newaxis], axis=-1)

    def reconstruct(self, compressed):
        return np.squeeze(self.projection_matrix @ compressed[..., np.newaxis], axis=-1)

    def latent_size(self):
        return self.rank

    def estimate_latent_ranges(self):
        if self.map_exists:
            latent_data = self.compress(self.data_matrix.T)  # (rank, num_samples)
            latent_min = np.min(latent_data, axis=0)
            latent_max = np.max(latent_data, axis=0)
            return [(lmin, lmax) for lmin, lmax in zip(latent_min, latent_max)]
        else:
            return None



"""Test interpolation classes. Currently, only barycentric Lagrange interpolation is supported."""
import itertools

import matplotlib.pyplot as plt
import numpy as np
from uqtils import approx_hess, approx_jac, ax_default

from amisc.examples.models import nonlinear_wave, tanh_func
from amisc.interpolator import Lagrange
from amisc.training import SparseGrid
from amisc.utils import relative_error
from amisc.variable import Variable


def test_tensor_product_1d(plots=False):
    """Test 1D tensor product Lagrange interpolation."""
    interp = Lagrange()
    x = Variable(distribution='U(0, 1)')
    alpha = ()          # no model fidelity for simple examples
    beta_interp = ()    # no beta for Lagrange interpolator
    domain = {'x': x.get_domain()}

    # Use sparse grid to get training data
    grid = SparseGrid()
    for beta in range(4):
        new_idx, new_x = grid.refine(alpha, (beta,), domain)
        new_y = tanh_func(new_x)
        grid.set(alpha, (beta,), new_idx, new_y)

    # Train the interpolator
    old_state = None
    training_data = grid.get(alpha, (3,))
    new_state = interp.refine(beta_interp, training_data, old_state, domain)

    x_grid = np.linspace(0, 1, 100)
    y_grid = tanh_func({'x': x_grid})['y']
    y_interp = interp.predict({'x': x_grid}, new_state, training_data)

    # Refine the sparse grid and predict again
    new_idx, new_x = grid.refine(alpha, (4,), domain)
    grid.set(alpha, (4,), new_idx, tanh_func(new_x))
    training_data = grid.get(alpha, (4,))
    new_state = interp.refine(beta_interp, training_data, new_state, domain)
    y2_interp = interp.predict({'x': x_grid}, new_state, training_data)

    # Compute errors
    N = 1000
    xtest = np.random.rand(N)
    ytest = interp.predict({'x': xtest}, new_state, training_data)['y']
    ytruth = tanh_func({'x': xtest})['y']
    l2_error = relative_error(ytest, ytruth)
    assert l2_error < 1e-1

    # Plot results
    if plots:
        fig, ax = plt.subplots()
        ax.plot(x_grid, y_grid, '-k', label='Model')
        xi, yi = grid.get(alpha, (3,))
        ax.plot(xi['x'], yi['y'], 'or', markersize=6, label=r'Training data')
        ax.plot(x_grid, y_interp['y'], '-r', label=r'$\beta=3$')
        xi, yi = grid.get(alpha, (4,))
        ax.plot(xi['x'], yi['y'], 'ob', markersize=4)
        ax.plot(x_grid, y2_interp['y'], '-b', label=r'$\beta=4$')
        ax_default(ax, r'Input', r'Output', legend=True)
        fig.tight_layout()
        plt.show()


def test_tensor_product_2d(plots=False):
    """Test 2D tensor product Lagrange interpolation."""
    bb_2d_func = lambda x: nonlinear_wave(x, env_var=0.2**2, wave_amp=0.3)
    interp = Lagrange()
    d = Variable(distribution='U(0, 1)')
    theta = Variable(distribution='U(0, 1)')
    alpha = ()          # no model fidelity for simple examples
    beta_interp = ()    # no beta for Lagrange interpolator
    domains = {'d': d.get_domain(), 'theta': theta.get_domain()}

    # Use sparse grid to get training data
    grid = SparseGrid()
    for beta in itertools.product(*[range(4) for _ in range(2)]):
        new_idx, new_x = grid.refine(alpha, beta, domains)
        new_y = bb_2d_func(new_x)
        grid.set(alpha, beta, new_idx, new_y)

    # Train the interpolator
    old_state = None
    training_data = grid.get(alpha, (3, 3))
    new_state = interp.refine(beta_interp, training_data, old_state, domains)

    # Predict
    N = 50
    x_grid = np.linspace(0, 1, N)
    xg, yg = np.meshgrid(x_grid, x_grid)
    zg = bb_2d_func({'d': xg, 'theta': yg})['y']
    z_interp = interp.predict({'d': xg, 'theta': yg}, new_state, training_data)['y']
    error = np.abs(z_interp - zg)

    # Refine interpolator
    new_beta = (3, 4)
    new_idx, new_x = grid.refine(alpha, new_beta, domains)
    grid.set(alpha, new_beta, new_idx, bb_2d_func(new_x))
    training_data = grid.get(alpha, new_beta)
    new_state = interp.refine(beta_interp, training_data, new_state, domains)
    z2_interp = interp.predict({'d': xg, 'theta': yg}, new_state, training_data)['y']
    error2 = np.abs(z2_interp - zg)

    vmin = min(np.min(z_interp), np.min(zg), np.min(z2_interp))
    vmax = max(np.max(z_interp), np.max(zg), np.max(z2_interp))
    emin = min(np.min(error), np.min(error2))
    emax = max(np.max(error), np.max(error2))

    # Compute errors
    N = 1000
    xtest = {'d': np.random.rand(N), 'theta': np.random.rand(N)}
    ytest = interp.predict(xtest, new_state, training_data)['y']
    ytruth = bb_2d_func(xtest)['y']
    l2_error = relative_error(ytest, ytruth)
    assert l2_error < 1e-1

    if plots:
        s = 100
        fig, ax = plt.subplots(2, 3)
        c1 = ax[0, 0].contourf(xg, yg, zg, 60, cmap='coolwarm', vmin=vmin, vmax=vmax)
        plt.colorbar(c1, ax=ax[0, 0])
        ax[0, 0].set_title('True function')
        ax_default(ax[0, 0], r'$x_1$', r'$x_2$', legend=False)
        c2 = ax[0, 1].contourf(xg, yg, z_interp, 60, cmap='coolwarm', vmin=vmin,vmax=vmax)
        xi, yi = grid.get(alpha, (3, 3))
        ax[0, 1].scatter(xi['d'], xi['theta'], c=yi['y'], marker='o', s=s, cmap='coolwarm', vmin=vmin, vmax=vmax,
                         alpha=1, linewidths=2, edgecolors='black')
        plt.colorbar(c2, ax=ax[0, 1])
        ax[0, 1].set_title('Interpolant')
        ax_default(ax[0, 1], r'$x_1$', '', legend=False)
        c3 = ax[0, 2].contourf(xg, yg, error, 60, cmap='viridis', vmin=emin, vmax=emax)
        ax[0, 2].plot(xi['d'], xi['theta'], 'o', markersize=6, markerfacecolor='green')
        plt.colorbar(c3, ax=ax[0, 2])
        ax[0, 2].set_title('Absolute error')
        ax_default(ax[0, 2], r'$x_1$', '', legend=False)
        c1 = ax[1, 0].contourf(xg, yg, zg, 60, cmap='coolwarm', vmin=vmin, vmax=vmax)
        plt.colorbar(c1, ax=ax[1, 0])
        ax[1, 0].set_title('True function')
        ax_default(ax[1, 0], r'$x_1$', r'$x_2$', legend=False)
        c2 = ax[1, 1].contourf(xg, yg, z2_interp, 60, cmap='coolwarm',vmin=vmin,vmax=vmax)
        xi, yi = grid.get(alpha, (3, 4))
        ax[1, 1].scatter(xi['d'], xi['theta'], c=yi['y'], marker='o', s=s, cmap='coolwarm', vmin=vmin, vmax=vmax,
                         alpha=1, linewidths=2, edgecolors='black')
        plt.colorbar(c2, ax=ax[1, 1])
        ax[1, 1].set_title('Refined')
        ax_default(ax[1, 1], r'$x_1$', '', legend=False)
        c3 = ax[1, 2].contourf(xg, yg, error2, 60, cmap='viridis', vmin=emin, vmax=emax)
        ax[1, 2].plot(xi['d'], xi['theta'], 'o', markersize=6, markerfacecolor='green')
        plt.colorbar(c3, ax=ax[1, 2])
        ax[1, 2].set_title('Absolute error')
        ax_default(ax[1, 2], r'$x_1$', '', legend=False)
        fig.set_size_inches(15, 11)
        fig.tight_layout()
        plt.show()


def test_interp_jacobian_and_hessian():
    f1 = lambda theta: 2 * theta['x1'] ** 2 + 3 * theta['x1'] * theta['x2'] ** 3 + np.cos(theta['x3'])
    f2 = lambda theta: (4 * theta['x1'] ** 2 + 2 * theta['x1'] ** 3 * theta['x2'] +
                        np.sin(theta['x3']) * theta['x1'])
    fun = lambda theta: {'y1': f1(theta), 'y2': f2(theta)}

    x1, x2, x3 = (Variable('x1', distribution='U(-2, 1)'), Variable('x2', distribution='U(-1, 2)'),
                  Variable('x3', distribution='U(-3.14, 3.14)'))
    interp = Lagrange()
    domains = {'x1': x1.get_domain(), 'x2': x2.get_domain(), 'x3': x3.get_domain()}
    alpha = ()
    beta_interp = ()
    beta_train = (3, 2, 4)

    # Use sparse grid to get training data
    grid = SparseGrid()
    for beta in itertools.product(*[range(b+1) for b in beta_train]):
        new_idx, new_x = grid.refine(alpha, beta, domains)
        new_y = fun(new_x)
        grid.set(alpha, beta, new_idx, new_y)

    old_state = None
    training_data = grid.get(alpha, beta_train)
    new_state = interp.refine(beta_interp, training_data, old_state, domains)

    N = (10, 11)

    def fun_vec(theta: np.ndarray):
        theta_dict = {var: theta[..., i] for i, var in enumerate(['x1', 'x2', 'x3'])}
        y_dict = fun(theta_dict)
        y_vec = np.concatenate([y_dict[var][..., np.newaxis] for var in ['y1', 'y2']], axis=-1)
        return y_vec

    xtest = {'x1': x1.sample(N), 'x2': x2.sample(N), 'x3': x3.sample(N)}
    xvec = np.concatenate([xtest[var][..., np.newaxis] for var in ['x1', 'x2', 'x3']], axis=-1)

    jac_truth = approx_jac(fun_vec, xvec)
    jac_interp = interp.gradient(xtest, new_state, training_data)
    jac_interp_vec = np.concatenate([np.expand_dims(jac_interp[var], axis=-2) for var in ['y1', 'y2']], axis=-2)
    assert np.allclose(jac_truth, jac_interp_vec, rtol=1e-2, atol=1e-2)

    hess_truth = approx_hess(fun_vec, xvec)
    hess_interp = interp.hessian(xtest, new_state, training_data)
    hess_interp_vec = np.concatenate([np.expand_dims(hess_interp[var], axis=-3) for var in ['y1', 'y2']], axis=-3)
    assert np.allclose(hess_truth, hess_interp_vec, rtol=1e-1, atol=1e-1)

    # Make sure interp gradient/hessian works directly at interpolation points as well
    xi, yi = training_data
    xtest = {'x1': x1.sample(10), 'x2': x2.sample(10), 'x3': x3.sample(10)}
    xi_test = {var: np.concatenate((xi[var], xtest[var]), axis=0) for var in ['x1', 'x2', 'x3']}
    N = xi_test['x1'].shape[0]
    idx = np.arange(0, N)
    np.random.shuffle(idx)
    for var in xi_test:
        xi_test[var] = xi_test[var][idx]
    xi_vec = np.concatenate([xi_test[var][..., np.newaxis] for var in ['x1', 'x2', 'x3']], axis=-1)

    jac_truth = approx_jac(fun_vec, xi_vec)
    jac_interp = interp.gradient(xi_test, new_state, training_data)
    jac_interp_vec = np.concatenate([np.expand_dims(jac_interp[var], axis=-2) for var in ['y1', 'y2']], axis=-2)
    assert np.allclose(jac_truth, jac_interp_vec, rtol=1e-2, atol=1e-2)

    hess_truth = approx_hess(fun_vec, xi_vec)
    hess_interp = interp.hessian(xi_test, new_state, training_data)
    hess_interp_vec = np.concatenate([np.expand_dims(hess_interp[var], axis=-3) for var in ['y1', 'y2']], axis=-3)
    assert np.allclose(hess_truth, hess_interp_vec, rtol=1e-1, atol=3e-1)



"""Provides example models and tutorials for the `amisc` package.

Includes:

- `models`: Example models for testing and demonstration.
- `tutorial`: Example tutorials for using the `amisc` package.
"""


"""Test that the [`SystemSurrogate`][amisc.system.SystemSurrogate] class validation works."""
# ruff: noqa: E741
import os
import shutil
import warnings
from pathlib import Path

import numpy as np
import pytest

from amisc import Component, System, Variable, YamlLoader
from amisc.system import TrainHistory


def simple_model(inputs, output_path='.'):
    return {'y': inputs['x'] ** 2}


def test_basic_init(tmp_path):
    """Test inputs/output/graph and root directory/logger/executor initialization."""
    comp1 = Component(simple_model, [Variable(l) for l in 'abc'], [Variable(l) for l in 'def'], name='comp1')
    comp2 = Component(simple_model, [Variable(l) for l in 'dghi'], [Variable(l) for l in 'jkl'], name='comp2')
    comp3 = Component(simple_model, [Variable(l) for l in 'lmno'], [Variable(l) for l in 'pqra'], name='comp3')
    system = System(comp1, [comp2], comp3)
    assert list(system.graph().edges()) == [('comp1', 'comp2'), ('comp2', 'comp3'), ('comp3', 'comp1')]
    assert system.inputs().keys() == {'b', 'c', 'g', 'h', 'i', 'm', 'n', 'o'}
    assert system.outputs().keys() == {'d', 'e', 'f', 'j', 'k', 'l', 'p', 'q', 'r', 'a'}

    paths = [tmp_path, tmp_path / 'amisc_test']
    for p in paths:
        s1 = System(comp1, comp2, comp3, root_dir=p)
        assert s1.root_dir.name.startswith('amisc_')
        assert (s1.root_dir / 'surrogates').is_dir()
        for comp in s1.components:
            assert (s1.root_dir / 'components' / comp.name).is_dir()
            assert comp.logger.handlers[0].baseFilename == s1.logger.handlers[0].baseFilename

    serialize_kwargs = {comp.name: {'training_data': {'save_path': tmp_path / f'{comp.name}_training_data.pkl'}}
                        for comp in s1.components}
    assert System.deserialize(s1.serialize(serialize_kwargs=serialize_kwargs)) == s1


def test_init_methods():
    """Test convenience methods for specifying a model."""
    # Call/return unpacked
    def my_model(x1, x2, model_fidelity=(1,)):
        y1 = x1 ** 2
        y2 = x2 ** 3
        return y1, y2
    s = System(my_model)
    inputs = {'x1': np.linspace(0, 1, 50), 'x2': np.linspace(0, 1, 50)}
    outputs = s.predict(inputs, use_model='best')
    expected = my_model(**inputs)
    assert np.allclose(outputs['y1'], expected[0])
    assert np.allclose(outputs['y2'], expected[1])
    assert s['my_model'].call_unpacked and s['my_model'].ret_unpacked
    assert all([v in s['my_model'].inputs for v in ['x1', 'x2']])
    assert all([v in s['my_model'].outputs for v in ['y1', 'y2']])

    # Call unpacked, return packed
    def my_model(x1, x2):
        return {'y1': x1 ** 2, 'y2': x2 ** 2}
    s = System(Component(my_model, Variable('x1', domain=(1, 4)), Variable('x2', domain=(0, 1)), 'y1', 'y2'))
    inputs = s.sample_inputs(10)
    outputs = s.predict(inputs, use_model='best')
    expected = my_model(inputs['x1'], inputs['x2'])
    assert np.allclose(outputs['y1'], expected['y1'])
    assert np.allclose(outputs['y2'], expected['y2'])
    assert s['my_model'].call_unpacked and not s['my_model'].ret_unpacked

    # Call packed, return unpacked
    def my_model(inputs):
        y1 = inputs['x1'] ** 2
        y2 = inputs['x2'] ** 2
        return y1, y2
    s = System(Component(my_model, [Variable('x1', domain=(-1, 1)), Variable('x2', domain=(-1, 1))]))
    inputs = s.sample_inputs((10, 2))
    outputs = s.predict(inputs, use_model='best')
    expected = my_model(inputs)
    assert np.allclose(outputs['y1'], expected[0])
    assert np.allclose(outputs['y2'], expected[1])
    assert not s['my_model'].call_unpacked and s['my_model'].ret_unpacked

    # Call/return packed
    def my_model(inputs):
        return {'y1': inputs['x1'] ** 2, 'y2': inputs['x2'] ** 2}
    s = System(Component(my_model, [Variable('x1', domain=(-1, 1)), Variable('x2', domain=(-1, 1))], ['y1', 'y2']))
    inputs = s.sample_inputs((20, 1))
    outputs = s.predict(inputs, use_model='best')
    expected = my_model(inputs)
    assert np.allclose(outputs['y1'], expected['y1'])
    assert np.allclose(outputs['y2'], expected['y2'])
    assert not s['my_model'].call_unpacked and not s['my_model'].ret_unpacked

    # Assert warning when trying unpacked usage of single input/output model
    def my_model(x1):
        y1 = x1 ** 2
        return y1
    with warnings.catch_warnings(record=True) as w:
        s = System(Component(my_model, Variable('x1'), Variable('y1')))
        assert len(w) == 2
        assert issubclass(w[0].category, UserWarning)
        assert issubclass(w[1].category, UserWarning)
        assert 'If you intended to use a single input argument' in str(w[0].message)
        assert 'If you intended to output a single return value' in str(w[1].message)
        assert not s['my_model'].call_unpacked and not s['my_model'].ret_unpacked

    # Assert value error raised if missing inputs or outputs
    def my_model(inputs):
        return {'y1': inputs['x1'] ** 2}
    with pytest.raises(ValueError):
        s = System(Component(my_model, inputs=None, outputs=None))
    with pytest.raises(ValueError):
        s2 = System(lambda: None)  # noqa: F841


def test_save_and_load(tmp_path):
    save_file = tmp_path / 'system.yml'
    s1 = System(Component(simple_model, Variable('x'), Variable('y'), name='comp1'))
    YamlLoader.dump(s1, save_file)
    os.mkdir(tmp_path / 'new_path')
    shutil.move(save_file, tmp_path / 'new_path' / save_file.name)
    save_file = tmp_path / 'new_path' / save_file.name
    for f in os.listdir(tmp_path):
        if Path(f).suffix == '.pkl':
            shutil.move(tmp_path / f, tmp_path / 'new_path' / f)
    os.chdir(tmp_path / 'new_path')  # change directory to test that yaml load still finds the .pkl save files
    s2 = YamlLoader.load(save_file)
    assert s2 == s1


def test_empty_inspection():
    """Load a completely empty system. Get variables and components from inspection."""
    def first(x):
        y = x ** 2
        return y

    def second(x, y):
        z = x + y
        z2 = y ** 2
        return z, z2
    system = System(first, second)
    assert all([comp.call_unpacked for comp in system.components])
    assert all([var in ['x'] for var in system.inputs()])
    assert all([var in ['y', 'z', 'z2'] for var in system.outputs()])
    assert system['first'].inputs['x'].name == 'x'
    assert system['first'].outputs['y'].name == 'y'
    assert system['first'].inputs['x'] is system['second'].inputs['x']
    assert system['first'].outputs['y'] is system['second'].inputs['y']

    x_grid = np.linspace(0, 1, 100)
    y_grid = system.predict({'x': x_grid}, use_model='best')
    assert np.allclose(first(x_grid), y_grid['y'])
    assert np.allclose(second(x_grid, first(x_grid))[0], y_grid['z'])
    assert np.allclose(second(x_grid, first(x_grid))[1], y_grid['z2'])


def test_swap_and_insert():
    """Test swapping and inserting a new component into a system."""
    def first(x):
        y = x ** 2
        return y

    def first_swap(x):
        y = x ** 3
        y2 = x ** 2 / 2
        return y, y2

    def second(x, y):
        z = x + y
        z2 = y ** 2
        return z, z2

    def third(z, z2):
        z3 = z + z2
        return z3

    system = System(first, second)
    system.insert_components(third)
    system.swap_component('first', first_swap)
    assert all([comp.name in ['first_swap', 'second', 'third'] for comp in system.components])
    assert all([var in system.inputs() for var in ['x']])
    assert all([var in system.outputs() for var in ['y', 'y2', 'z', 'z2', 'z3']])

    system.remove_component('third')
    assert len(system.components) == 2
    assert system.components[0].name == 'first_swap'
    assert system.components[1].name == 'second'


def test_train_history():
    # Test TrainHistory data structure
    hist = [{'component': f'comp{i}', 'alpha': tuple(np.random.randint(0, 10, 3)),
             'beta': tuple(np.random.randint(0, 10, 2)), 'num_evals': np.random.rand(),
             'added_cost': np.random.rand(), 'added_error': np.random.rand(), 'test_error': {'y': np.random.rand()}}
            for i in range(10)]
    new_hist = [{'component': f'comp{int(np.random.randint(10))}', 'alpha': tuple(np.random.randint(0, 3, 2)),
                 'beta': tuple(np.random.randint(0, 5, 4)), 'num_evals': np.random.rand(),
                 'added_cost': np.random.rand(), 'added_error': np.random.rand()} for i in range(5)]
    hist = TrainHistory(data=hist)
    hist.extend(new_hist)
    assert TrainHistory.deserialize(hist.serialize()) == hist


def test_save_dir(tmp_path):
    """Test saving and loading from `amisc_timestamp` directory"""
    comp = Component(simple_model, Variable('x', domain=(0, 1)), Variable('y'), name='comp1', data_fidelity=(4,))
    surr = System(comp, root_dir=tmp_path, name='test_system')
    surr.fit(max_iter=3, save_interval=1)
    surr.save_to_file('test_surrogate.yml')

    found_dir = False
    for f in os.listdir(tmp_path):
        if f.startswith('amisc_'):
            amisc_dir = tmp_path / f
            surr2 = System.load_from_file(amisc_dir / 'surrogates' / 'test_surrogate.yml')
            assert (amisc_dir / 'components' / 'comp1').is_dir()
            assert all([(amisc_dir / 'surrogates' / f'{surr.name}_iter{i}' / f'{surr.name}_iter{i}.yml').exists()
                        for i in range(1, 4)])
            assert surr2 == surr
            found_dir = True

    assert found_dir



"""A `Component` is an `amisc` wrapper around a single discipline model. It manages surrogate construction and
a hierarchy of modeling fidelities.

!!! Info "Multi-indices in the MISC approximation"
    A multi-index is a tuple of natural numbers, each specifying a level of fidelity. You will frequently see two
    multi-indices: `alpha` and `beta`. The `alpha` (or $\\alpha$) indices specify physical model fidelity and get
    passed to the model as an additional argument (e.g. things like discretization level, time step size, etc.).
    The `beta` (or $\\beta$) indices specify surrogate refinement level, so typically an indication of the amount of
    training data used or the complexity of the surrogate model. We divide $\\beta$ into `data_fidelity` and
    `surrogate_fidelity` for specifying training data and surrogate model complexity, respectively.

Includes:

- `ModelKwargs` — a dataclass for storing model keyword arguments
- `StringKwargs` — a dataclass for storing model keyword arguments as a string
- `IndexSet` — a dataclass that maintains a list of multi-indices
- `MiscTree` — a dataclass that maintains MISC data in a `dict` tree, indexed by `alpha` and `beta`
- `Component` — a class that manages a single discipline model and its surrogate hierarchy
"""
from __future__ import annotations

import ast
import copy
import inspect
import itertools
import logging
import random
import string
import traceback
import typing
import warnings
from collections import UserDict, deque
from concurrent.futures import ALL_COMPLETED, Executor, wait
from pathlib import Path
from typing import Any, Callable, ClassVar, Iterable, Literal, Optional

import numpy as np
import yaml
from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator
from typing_extensions import TypedDict

from amisc.interpolator import Interpolator, InterpolatorState, Lagrange
from amisc.serialize import PickleSerializable, Serializable, StringSerializable, YamlSerializable
from amisc.training import SparseGrid, TrainingData
from amisc.typing import LATENT_STR_ID, Dataset, MultiIndex
from amisc.utils import (
    _get_yaml_path,
    _inspect_assignment,
    _inspect_function,
    format_inputs,
    format_outputs,
    get_logger,
    search_for_file,
    to_model_dataset,
    to_surrogate_dataset,
)
from amisc.variable import Variable, VariableList

__all__ = ["ModelKwargs", "StringKwargs", "IndexSet", "MiscTree", "Component"]
_VariableLike = list[Variable | dict | str] | str | Variable | dict | VariableList  # Generic type for Variables


class ModelKwargs(UserDict, Serializable):
    """Default dataclass for storing model keyword arguments in a `dict`. If you have kwargs that require
    more complicated serialization/specification than a plain `dict`, then you can subclass from here.
    """

    def serialize(self):
        return self.data

    @classmethod
    def deserialize(cls, serialized_data):
        return ModelKwargs(**serialized_data)

    @classmethod
    def from_dict(cls, config: dict) -> ModelKwargs:
        """Create a `ModelKwargs` object from a `dict` configuration."""
        method = config.pop('method', 'default_kwargs').lower()
        match method:
            case 'default_kwargs':
                return ModelKwargs(**config)
            case 'string_kwargs':
                return StringKwargs(**config)
            case other:
                config['method'] = other
                return ModelKwargs(**config)  # Pass the method through


class StringKwargs(StringSerializable, ModelKwargs):
    """Dataclass for storing model keyword arguments as a string."""
    def __repr__(self):
        return str(self.data)

    def __str__(self):
        def format_value(value):
            if isinstance(value, str):
                return f'"{value}"'
            else:
                return str(value)

        kw_str = ", ".join([f"{key}={format_value(value)}" for key, value in self.items()])
        return f"ModelKwargs({kw_str})"


class IndexSet(set, Serializable):
    """Dataclass that maintains a list of multi-indices. Overrides basic `set` functionality to ensure
    elements are formatted correctly as `(alpha, beta)`; that is, as a tuple of `alpha` and
    `beta`, which are themselves instances of a [`MultiIndex`][amisc.typing.MultiIndex] tuple.

    !!! Example "An example index set"
        $\\mathcal{I} = [(\\alpha, \\beta)_1 , (\\alpha, \\beta)_2, (\\alpha, \\beta)_3 , ...]$ would be specified
        as `I = [((0, 0), (0, 0, 0)) , ((0, 1), (0, 1, 0)), ...]`.
    """
    def __init__(self, s=()):
        s = [self._validate_element(ele) for ele in s]
        super().__init__(s)

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return self.__str__()

    def add(self, __element):
        super().add(self._validate_element(__element))

    def update(self, __elements):
        super().update([self._validate_element(ele) for ele in __elements])

    @classmethod
    def _validate_element(cls, element):
        """Validate that the element is a tuple of two multi-indices."""
        alpha, beta = ast.literal_eval(element) if isinstance(element, str) else tuple(element)
        return MultiIndex(alpha), MultiIndex(beta)

    @classmethod
    def _wrap_methods(cls, names):
        """Make sure set operations return an `IndexSet` object."""
        def wrap_method_closure(name):
            def inner(self, *args):
                result = getattr(super(cls, self), name)(*args)
                if isinstance(result, set):
                    result = cls(result)
                return result
            inner.fn_name = name
            setattr(cls, name, inner)

        for name in names:
            wrap_method_closure(name)

    def serialize(self) -> list[str]:
        """Return a list of each multi-index in the set serialized to a string."""
        return [str(ele) for ele in self]

    @classmethod
    def deserialize(cls, serialized_data: list[str]) -> IndexSet:
        """Deserialize a list of tuples to an `IndexSet`."""
        return cls(serialized_data)


IndexSet._wrap_methods(['__ror__', 'difference_update', '__isub__', 'symmetric_difference', '__rsub__', '__and__',
                        '__rand__', 'intersection', 'difference', '__iand__', 'union', '__ixor__',
                        'symmetric_difference_update', '__or__', 'copy', '__rxor__', 'intersection_update', '__xor__',
                        '__ior__', '__sub__'
                        ])


class MiscTree(UserDict, Serializable):
    """Dataclass that maintains MISC data in a `dict` tree, indexed by `alpha` and `beta`. Overrides
    basic `dict` functionality to ensure elements are formatted correctly as `(alpha, beta) -> data`.
    Used to store MISC coefficients, model costs, and interpolator states.

    The underlying data structure is: `dict[MultiIndex, dict[MultiIndex, float | InterpolatorState]]`.
    """
    SERIALIZER_KEY = 'state_serializer'

    def __init__(self, data: dict = None, **kwargs):
        data_dict = data or {}
        if isinstance(data_dict, MiscTree):
            data_dict = data_dict.data
        data_dict.update(kwargs)
        super().__init__(self._validate_data(data_dict))

    def serialize(self, *args, keep_yaml_objects=False, **kwargs) -> dict:
        """Serialize `alpha, beta` indices to string and return a `dict` of internal data.

        :param args: extra serialization arguments for internal `InterpolatorState`
        :param keep_yaml_objects: whether to keep `YamlSerializable` instances in the serialization
        :param kwargs: extra serialization keyword arguments for internal `InterpolatorState`
        """
        ret_dict = {}
        if state_serializer := self.state_serializer(self.data):
            ret_dict[self.SERIALIZER_KEY] = state_serializer.obj if keep_yaml_objects else state_serializer.serialize()
        for alpha, beta, data in self:
            ret_dict.setdefault(str(alpha), dict())
            serialized_data = data.serialize(*args, **kwargs) if isinstance(data, InterpolatorState) else float(data)
            ret_dict[str(alpha)][str(beta)] = serialized_data
        return ret_dict

    @classmethod
    def deserialize(cls, serialized_data: dict) -> MiscTree:
        """Deserialize a `dict` to a `MiscTree`.

        :param serialized_data: the data to deserialize to a `MiscTree` object
        """
        return cls(serialized_data)

    @classmethod
    def state_serializer(cls, data: dict) -> YamlSerializable | None:
        """Infer and return the interpolator state serializer from the `MiscTree` data (if possible). If no
        `InterpolatorState` instance could be found, return `None`.
        """
        serializer = data.get(cls.SERIALIZER_KEY, None)  # if `data` is serialized
        if serializer is None:  # Otherwise search for an InterpolatorState
            for alpha, beta_dict in data.items():
                if alpha == cls.SERIALIZER_KEY:
                    continue
                for beta, value in beta_dict.items():
                    if isinstance(value, InterpolatorState):
                        serializer = type(value)
                        break
                if serializer is not None:
                    break
        return cls._validate_state_serializer(serializer)

    @classmethod
    def _validate_state_serializer(cls, state_serializer: Optional[str | type[Serializable] | YamlSerializable]
                                   ) -> YamlSerializable | None:
        if state_serializer is None:
            return None
        elif isinstance(state_serializer, YamlSerializable):
            return state_serializer
        elif isinstance(state_serializer, str):
            return YamlSerializable.deserialize(state_serializer)  # Load the serializer type from string
        else:
            return YamlSerializable(obj=state_serializer)

    @classmethod
    def _validate_data(cls, serialized_data: dict) -> dict:
        state_serializer = cls.state_serializer(serialized_data)
        ret_dict = {}
        for alpha, beta_dict in serialized_data.items():
            if alpha == cls.SERIALIZER_KEY:
                continue
            alpha_tup = MultiIndex(alpha)
            ret_dict.setdefault(alpha_tup, dict())
            for beta, data in beta_dict.items():
                beta_tup = MultiIndex(beta)
                if isinstance(data, InterpolatorState):
                    pass
                elif state_serializer is not None:
                    data = state_serializer.obj.deserialize(data)
                else:
                    data = float(data)
                assert isinstance(data, InterpolatorState | float)
                ret_dict[alpha_tup][beta_tup] = data
        return ret_dict

    @staticmethod
    def _is_alpha_beta_access(key):
        """Check that the key is of the format `(alpha, beta).`"""
        return (isinstance(key, tuple) and len(key) == 2 and isinstance(key[0], str | tuple)
                and isinstance(key[1], str | tuple))

    def get(self, key, default=None) -> float | InterpolatorState:
        try:
            return self.__getitem__(key)
        except Exception:
            return default

    def update(self, data_dict: dict = None, **kwargs):
        """Force `dict.update()` through the validator."""
        data_dict = data_dict or dict()
        data_dict.update(kwargs)
        super().update(self._validate_data(data_dict))

    def __setitem__(self, key: tuple | MultiIndex, value: float | InterpolatorState):
        """Allows `misc_tree[alpha, beta] = value` usage."""
        if self._is_alpha_beta_access(key):
            alpha, beta = MultiIndex(key[0]), MultiIndex(key[1])
            self.data.setdefault(alpha, dict())
            self.data[alpha][beta] = value
        else:
            super().__setitem__(MultiIndex(key), value)

    def __getitem__(self, key: tuple | MultiIndex) -> float | InterpolatorState:
        """Allows `value = misc_tree[alpha, beta]` usage."""
        if self._is_alpha_beta_access(key):
            alpha, beta = MultiIndex(key[0]), MultiIndex(key[1])
            return self.data[alpha][beta]
        else:
            return super().__getitem__(MultiIndex(key))

    def clear(self):
        """Clear the `MiscTree` data."""
        for key in list(self.data.keys()):
            del self.data[key]

    def __eq__(self, other):
        if isinstance(other, MiscTree):
            try:
                for alpha, beta, data in self:
                    if other[alpha, beta] != data:
                        return False
                return True
            except KeyError:
                return False
        else:
            return False

    def __iter__(self) -> Iterable[tuple[tuple, tuple, float | InterpolatorState]]:
        for alpha, beta_dict in self.data.items():
            if alpha == self.SERIALIZER_KEY:
                continue
            for beta, data in beta_dict.items():
                yield alpha, beta, data


class ComponentSerializers(TypedDict, total=False):
    """Type hint for the `Component` class data serializers.

    :ivar model_kwargs: the model kwarg object class
    :ivar interpolator: the interpolator object class
    :ivar training_data: the training data object class
    """
    model_kwargs: str | type[Serializable] | YamlSerializable
    interpolator: str | type[Serializable] | YamlSerializable
    training_data: str | type[Serializable] | YamlSerializable


class Component(BaseModel, Serializable):
    """A `Component` wrapper around a single discipline model. It manages MISC surrogate construction and a hierarchy of
    modeling fidelities.

    A `Component` can be constructed by specifying a model, input and output variables, and additional configurations
    such as the maximum fidelity levels, the interpolator type, and the training data type. If `model_fidelity`,
    `data_fidelity`, and `surrogate_fidelity` are all left empty, then the `Component` will not use a surrogate model,
    instead calling the underlying model directly. The `Component` can be serialized to a YAML file and deserialized
    back into a Python object.

    !!! Example "A simple `Component`"
        ```python
        from amisc import Component, Variable

        x = Variable(domain=(0, 1))
        y = Variable()
        model = lambda x: {'y': x['x']**2}
        comp = Component(model=model, inputs=[x], outputs=[y])
        ```

    Each fidelity index in $\\alpha$ increases in refinement from $0$ up to `model_fidelity`. Each fidelity index
    in $\\beta$ increases from $0$ up to `(data_fidelity, surrogate_fidelity)`. From the `Component's` perspective,
    the concatenation of $(\\alpha, \\beta)$ fully specifies a single fidelity "level". The `Component`
    forms an approximation of the model by summing up over many of these concatenated sets of $(\\alpha, \\beta)$.

    :ivar name: the name of the `Component`
    :ivar model: the model or function that is to be approximated, callable as `y = f(x)`
    :ivar inputs: the input variables to the model
    :ivar outputs: the output variables from the model
    :ivar model_kwargs: extra keyword arguments to pass to the model
    :ivar model_fidelity: the maximum level of refinement for each fidelity index in $\\alpha$ for model fidelity
    :ivar data_fidelity: the maximum level of refinement for each fidelity index in $\\beta$ for training data
    :ivar surrogate_fidelity: the max level of refinement for each fidelity index in $\\beta$ for the surrogate
    :ivar interpolator: the interpolator to use as the underlying surrogate model
    :ivar vectorized: whether the model supports vectorized input/output (i.e. datasets with arbitrary shape `(...,)`)
    :ivar call_unpacked: whether the model expects unpacked input arguments (i.e. `func(x1, x2, ...)`)
    :ivar ret_unpacked: whether the model returns unpacked output arguments (i.e. `func() -> (y1, y2, ...)`)

    :ivar active_set: the current active set of multi-indices in the MISC approximation
    :ivar candidate_set: all neighboring multi-indices that are candidates for inclusion in `active_set`
    :ivar misc_states: the interpolator states for each multi-index in the MISC approximation
    :ivar misc_costs: the computational cost associated with each multi-index in the MISC approximation
    :ivar misc_coeff_train: the combination technique coefficients for the active set multi-indices
    :ivar misc_coeff_test: the combination technique coefficients for the active and candidate set multi-indices
    :ivar model_costs: the average single fidelity model costs for each $\\alpha$
    :ivar training_data: the training data storage structure for the surrogate model

    :ivar serializers: the custom serializers for the `[model_kwargs, interpolator, training_data]`
                       `Component` attributes -- these should be the _types_ of the serializer objects, which will
                       be inferred from the data passed in if not explicitly set
    :ivar _logger: the logger for the `Component`
    """
    yaml_tag: ClassVar[str] = u'!Component'
    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True, validate_default=True,
                              protected_namespaces=(), extra='allow')
    # Configuration
    serializers: Optional[ComponentSerializers] = None
    name: Optional[str] = None
    model: str | Callable[[dict | Dataset, ...], dict | Dataset]
    model_kwargs: str | dict | ModelKwargs = {}
    inputs: _VariableLike
    outputs: _VariableLike
    model_fidelity: str | tuple = MultiIndex()
    data_fidelity: str | tuple = MultiIndex()
    surrogate_fidelity: str | tuple = MultiIndex()
    interpolator: Any | Interpolator = Lagrange()
    vectorized: bool = False
    call_unpacked: Optional[bool] = None  # If the model expects inputs/outputs like `func(x1, x2, ...)->(y1, y2, ...)
    ret_unpacked: Optional[bool] = None

    # Data storage/states for a MISC component
    active_set: list | set | IndexSet = IndexSet()     # set of active (alpha, beta) multi-indices
    candidate_set: list | set | IndexSet = IndexSet()  # set of candidate (alpha, beta) multi-indices
    misc_states: dict | MiscTree = MiscTree()          # (alpha, beta) -> Interpolator state
    misc_costs: dict | MiscTree = MiscTree()           # (alpha, beta) -> Added computational cost for this mult-index
    misc_coeff_train: dict | MiscTree = MiscTree()     # (alpha, beta) -> c_[alpha, beta] (active set only)
    misc_coeff_test: dict | MiscTree = MiscTree()      # (alpha, beta) -> c_[alpha, beta] (including candidate set)
    model_costs: dict = dict()                         # Average single fidelity model costs (for each alpha)
    training_data: Any | TrainingData = SparseGrid()   # Stores surrogate training data

    # Internal
    _logger: Optional[logging.Logger] = None

    def __init__(self, /, model, *args, inputs=None, outputs=None, name=None, **kwargs):
        if name is None:
            name = _inspect_assignment('Component')  # try to assign the name from inspection
        name = name or model.__name__ or "Component_" + "".join(random.choices(string.digits, k=3))

        # Determine how the model expects to be called and gather inputs/outputs
        _ = self._validate_model_signature(model, args, inputs, outputs, kwargs.get('call_unpacked', None),
                                           kwargs.get('ret_unpacked', None))
        model, inputs, outputs, call_unpacked, ret_unpacked = _
        kwargs['call_unpacked'] = call_unpacked
        kwargs['ret_unpacked'] = ret_unpacked

        # Gather all model kwargs (anything else passed in for kwargs is assumed to be a model kwarg)
        model_kwargs = kwargs.get('model_kwargs', {})
        for key in kwargs.keys() - self.model_fields.keys():
            model_kwargs[key] = kwargs.pop(key)
        kwargs['model_kwargs'] = model_kwargs

        # Gather data serializers from type checks (if not passed in as a kwarg)
        serializers = kwargs.get('serializers', {})  # directly passing serializers will override type checks
        for key in ComponentSerializers.__annotations__.keys():
            field = kwargs.get(key, None)
            if isinstance(field, dict):
                field_super = next(filter(lambda x: issubclass(x, Serializable),
                                          typing.get_args(self.model_fields[key].annotation)), None)
                field = field_super.from_dict(field) if field_super is not None else field
            if not serializers.get(key, None):
                serializers[key] = type(field) if isinstance(field, Serializable) else (
                    type(self.model_fields[key].default))
        kwargs['serializers'] = serializers

        super().__init__(model=model, inputs=inputs, outputs=outputs, name=name, **kwargs)  # Runs pydantic validation

        # Set internal properties
        assert self.is_downward_closed(self.active_set.union(self.candidate_set))
        self.set_logger()

    @classmethod
    def _validate_model_signature(cls, model, args=(), inputs=None, outputs=None,
                                  call_unpacked=None, ret_unpacked=None):
        """Parse model signature and decide how the model expects to be called based on what input/output information
        is provided or inspected from the model signature.
        """
        if inputs is not None:
            inputs = cls._validate_variables(inputs)
        if outputs is not None:
            outputs = cls._validate_variables(outputs)
        model = cls._validate_model(model)

        # Default to `dict` (i.e. packed) model call/return signatures
        if call_unpacked is None:
            call_unpacked = False
        if ret_unpacked is None:
            ret_unpacked = False
        inputs_inspect, outputs_inspect = _inspect_function(model)
        call_unpacked = call_unpacked or (len(inputs_inspect) > 1)  # Assume multiple inputs require unpacking
        ret_unpacked = ret_unpacked or (len(outputs_inspect) > 1)   # Assume multiple outputs require unpacking

        # Extract inputs/outputs from args
        arg_inputs = ()
        arg_outputs = ()
        if len(args) > 0:
            if call_unpacked:
                if isinstance(args[0], dict | str | Variable):
                    arg_inputs = args[:len(inputs_inspect)]
                    arg_outputs = args[len(inputs_inspect):]
                else:
                    arg_inputs = args[0]
                    arg_outputs = args[1:]
            else:
                arg_inputs = args[0]    # Assume first arg is a single or list of inputs
                arg_outputs = args[1:]  # Assume rest are outputs

        # Resolve inputs
        inputs = inputs or []
        inputs = VariableList.merge(inputs, arg_inputs)
        if len(inputs) == 0:
            inputs = inputs_inspect
            call_unpacked = True
            if len(inputs) == 0:
                raise ValueError("Could not infer input variables from model signature. Either your model does not "
                                 "accept input arguments or an error occurred during inspection.\nPlease provide the "
                                 "inputs directly as `Component(inputs=[...])` or fix the model signature.")
        if call_unpacked:
            if not all([var == inputs_inspect[i] for i, var in enumerate(inputs)]):
                warnings.warn(f"Mismatch between provided inputs: {inputs.values()} and inputs inferred from "
                              f"model signature: {inputs_inspect}. This may cause unexpected results.")
        else:
            if len(inputs_inspect) > 1:
                warnings.warn(f"Model signature expects multiple input arguments: {inputs_inspect}. "
                              f"Please set `call_unpacked=True` to use this model signature for multiple "
                              f"inputs.\nOtherwise, move all inputs into a single `dict` argument and all "
                              f"extra arguments into the `model_kwargs` field.")

            # Can't assume unpacked for single input/output, so warn user if they may be trying to do so
            if len(inputs) == 1 and len(inputs_inspect) == 1 and str(inputs[0]) == str(inputs_inspect[0]):
                warnings.warn(f"Single input argument: {inputs[0]} provided to model with input signature: "
                              f"{inputs_inspect}.\nIf you intended to use a single input argument, set "
                              f"`call_unpacked=True` to use this model signature.\nOtherwise, the first input will "
                              f"be passed to your model as a `dict`.\nIf you are expecting a `dict` input already, "
                              f"change the name of the input to not exactly "
                              f"match {inputs_inspect} in order to silence this warning.")
        # Resolve outputs
        outputs = outputs or []
        outputs = VariableList.merge(outputs, *arg_outputs)
        if len(outputs) == 0:
            outputs = outputs_inspect
            ret_unpacked = True
            if len(outputs) == 0:
                raise ValueError("Could not infer output variables from model inspection. Either your model does not "
                                 "return outputs or an error occurred during inspection.\nPlease provide the "
                                 "outputs directly as `Component(outputs=[...])` or fix the model return values.")
        if ret_unpacked:
            if not all([var == outputs_inspect[i] for i, var in enumerate(outputs)]):
                warnings.warn(f"Mismatch between provided outputs: {outputs.values()} and outputs inferred "
                              f"from model: {outputs_inspect}. This may cause unexpected results.")
        else:
            if len(outputs_inspect) > 1:
                warnings.warn(f"Model expects multiple return values: {outputs_inspect}. Please set "
                              f"`ret_unpacked=True` to use this model signature for multiple outputs.\n"
                              f"Otherwise, move all outputs into a single `dict` return value.")

            if len(outputs) == 1 and len(outputs_inspect) == 1 and str(outputs[0]) == str(outputs_inspect[0]):
                warnings.warn(f"Single output: {outputs[0]} provided to model with single expected return: "
                              f"{outputs_inspect}.\nIf you intended to output a single return value, set "
                              f"`ret_unpacked=True` to use this model signature.\nOtherwise, the output should "
                              f"be returned from your model as a `dict`.\nIf you are returning a `dict` already, "
                              f"then change its name to not exactly match {outputs_inspect} in order to silence "
                              f"this warning.")
        return model, inputs, outputs, call_unpacked, ret_unpacked

    def __repr__(self):
        s = f'---- {self.name} ----\n'
        s += f'Inputs:  {self.inputs}\n'
        s += f'Outputs: {self.outputs}\n'
        s += f'Model:   {self.model}'
        return s

    def __str__(self):
        return self.__repr__()

    @field_validator('serializers')
    @classmethod
    def _validate_serializers(cls, serializers: ComponentSerializers) -> ComponentSerializers:
        """Make sure custom serializer object types are themselves serializable as `YamlSerializable`."""
        for key, serializer in serializers.items():
            if serializer is None:
                serializers[key] = None
            elif isinstance(serializer, YamlSerializable):
                serializers[key] = serializer
            elif isinstance(serializer, str):
                serializers[key] = YamlSerializable.deserialize(serializer)
            else:
                serializers[key] = YamlSerializable(obj=serializer)
        return serializers

    @field_validator('model')
    @classmethod
    def _validate_model(cls, model: str | Callable) -> Callable:
        """Expects model as a callable or a yaml !!python/name string representation."""
        if isinstance(model, str):
            return YamlSerializable.deserialize(model).obj
        else:
            return model

    @field_validator('inputs', 'outputs')
    @classmethod
    def _validate_variables(cls, variables: _VariableLike) -> VariableList:
        if isinstance(variables, VariableList):
            return variables
        else:
            return VariableList.deserialize(variables)

    @field_validator('model_fidelity', 'data_fidelity', 'surrogate_fidelity')
    @classmethod
    def _validate_indices(cls, multi_index) -> MultiIndex:
        return MultiIndex(multi_index)

    @field_validator('active_set', 'candidate_set')
    @classmethod
    def _validate_index_set(cls, index_set) -> IndexSet:
        return IndexSet.deserialize(index_set)

    @field_validator('misc_states', 'misc_costs', 'misc_coeff_train', 'misc_coeff_test')
    @classmethod
    def _validate_misc_tree(cls, misc_tree) -> MiscTree:
        return MiscTree.deserialize(misc_tree)

    @field_validator('model_costs')
    @classmethod
    def _validate_model_costs(cls, model_costs: dict) -> dict:
        return {MultiIndex(key): float(value) for key, value in model_costs.items()}

    @field_validator('model_kwargs', 'interpolator', 'training_data')
    @classmethod
    def _validate_arbitrary_serializable(cls, data: Any, info: ValidationInfo) -> Any:
        """Use the stored custom serialization classes to deserialize arbitrary objects."""
        serializer = info.data.get('serializers').get(info.field_name).obj
        if isinstance(data, Serializable):
            return data
        else:
            return serializer.deserialize(data)

    @property
    def xdim(self) -> int:
        return len(self.inputs)

    @property
    def ydim(self) -> int:
        return len(self.outputs)

    @property
    def max_alpha(self) -> MultiIndex:
        """The maximum model fidelity multi-index (alias for `model_fidelity`)."""
        return self.model_fidelity

    @property
    def max_beta(self) -> MultiIndex:
        """The maximum surrogate fidelity multi-index is a combination of training and interpolator indices."""
        return self.data_fidelity + self.surrogate_fidelity

    @property
    def has_surrogate(self) -> bool:
        """The component has no surrogate model if there are no fidelity indices."""
        return (len(self.max_alpha) + len(self.max_beta)) > 0

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        self._logger = logger

    def __eq__(self, other):
        if isinstance(other, Component):
            return (self.model.__code__.co_code == other.model.__code__.co_code and self.inputs == other.inputs
                    and self.outputs == other.outputs and self.name == other.name
                    and self.model_kwargs.data == other.model_kwargs.data
                    and self.model_fidelity == other.model_fidelity and self.max_beta == other.max_beta and
                    self.interpolator == other.interpolator
                    and self.active_set == other.active_set and self.candidate_set == other.candidate_set
                    and self.misc_states == other.misc_states and self.misc_costs == other.misc_costs
                    )
        else:
            return False

    def _neighbors(self, alpha: MultiIndex, beta: MultiIndex, active_set: IndexSet = None, forward: bool = True):
        """Get all possible forward or backward multi-index neighbors (distance of one unit vector away).

        :param alpha: the model fidelity index
        :param beta: the surrogate fidelity index
        :param active_set: the set of active multi-indices
        :param forward: whether to get forward or backward neighbors
        :returns: a set of multi-indices that are neighbors of the input multi-index pair `(alpha, beta)`
        """
        active_set = active_set or self.active_set
        ind = np.array(alpha + beta)
        max_ind = np.array(self.max_alpha + self.max_beta)
        new_candidates = IndexSet()
        for i in range(len(ind)):
            ind_new = ind.copy()
            ind_new[i] += 1 if forward else -1

            # Don't add if we surpass a refinement limit or lower bound
            if np.any(ind_new > max_ind) or np.any(ind_new < 0):
                continue

            # Add the new index if it maintains downward-closedness
            down_closed = True
            for j in range(len(ind)):
                ind_check = ind_new.copy()
                ind_check[j] -= 1
                if ind_check[j] >= 0:
                    tup_check = (MultiIndex(ind_check[:len(alpha)]), MultiIndex(ind_check[len(alpha):]))
                    if tup_check not in active_set and tup_check != (alpha, beta):
                        down_closed = False
                        break
            if down_closed:
                new_candidates.add((ind_new[:len(alpha)], ind_new[len(alpha):]))

        return new_candidates

    def _surrogate_outputs(self):
        """Helper function to get the names of the surrogate outputs (including latent variables)."""
        y_vars = []
        for var in self.outputs:
            if var.compression is not None:
                for i in range(var.compression.latent_size()):
                    y_vars.append(f'{var.name}{LATENT_STR_ID}{i}')
            else:
                y_vars.append(var.name)
        return y_vars

    def _match_index_set(self, index_set, misc_coeff):
        """Helper function to grab the correct data structures for the given index set and MISC coefficients."""
        if misc_coeff is None:
            match index_set:
                case 'train':
                    misc_coeff = self.misc_coeff_train
                case 'test':
                    misc_coeff = self.misc_coeff_test
                case other:
                    raise ValueError(f"Index set must be 'train' or 'test' if you do not provide `misc_coeff`. "
                                     f"{other} not recognized.")
        if isinstance(index_set, str):
            match index_set:
                case 'train':
                    index_set = self.active_set
                case 'test':
                    index_set = self.active_set.union(self.candidate_set)
                case other:
                    raise ValueError(f"Index set must be 'train' or 'test'. {other} not recognized.")

        return index_set, misc_coeff

    def get_training_data(self, alpha: Literal['best', 'worst'] | MultiIndex = 'best',
                          beta: Literal['best', 'worst'] | MultiIndex = 'best',
                          y_vars: list = None) -> tuple[Dataset, Dataset]:
        """Get all training data for a given multi-index pair `(alpha, beta)`.

        :param alpha: the model fidelity index (defaults to the maximum available model fidelity)
        :param beta: the surrogate fidelity index (defaults to the maximum available surrogate fidelity)
        :param y_vars: the training data to return (defaults to all stored data)
        :returns: `(xtrain, ytrain)` - the training data for the given multi-indices
        """
        # Find the best alpha
        if alpha == 'best':
            alpha_best = ()
            for a, _ in self.active_set.union(self.candidate_set):
                if sum(a) > sum(alpha_best):
                    alpha_best = a
            alpha = alpha_best
        elif alpha == 'worst':
            alpha = (0,) * len(self.max_alpha)

        # Find the best beta for the given alpha
        if beta == 'best':
            beta_best = ()
            for a, b in self.active_set.union(self.candidate_set):
                if a == alpha and sum(b) > sum(beta_best):
                    beta_best = b
            beta = beta_best
        elif beta == 'worst':
            beta = (0,) * len(self.max_beta)

        try:
            return self.training_data.get(alpha, beta, y_vars=y_vars, skip_nan=True)
        except Exception as e:
            self.logger.error(f"Error getting training data for alpha={alpha}, beta={beta}.")
            raise e

    def call_model(self, inputs: dict | Dataset,
                   model_fidelity: Literal['best', 'worst'] | tuple | list = None,
                   output_path: str | Path = None,
                   executor: Executor = None,
                   **kwds) -> Dataset:
        """Wrapper function for calling the underlying component model.

        This function formats the input data, calls the model, and processes the output data.
        It supports vectorized calls, parallel execution using an executor, or serial execution. These options are
        checked in that order, with the first available method used. Must set `Component.vectorized=True` if the
        model supports input arrays of the form `(N,)` or even arbitrary shape `(...,)`.

        !!! Warning "Parallel Execution"
            The underlying model must be defined in a global module scope if `pickle` is the serialization method for
            the provided `Executor`.

        !!! Note "Additional return values"
            The model can return additional items that are not part of `Component.outputs`. These items are returned
            as object arrays in the output `dict`. Two special return values are `model_cost` and `output_path`.
            Returning `model_cost` will store the computational cost of a single model evaluation (which is used by
            `amisc` adaptive surrogate training). Returning `output_path` will store the output file name if the model
            wrote any files to disk.

        !!! Note "Handling errors"
            If the underlying component model raises an exception, the error is stored in `output_dict['errors']` with
            the index of the input data that caused the error. The output data for that index is set to `np.nan`
            for each output variable.

        :param inputs: The input data for the model, formatted as a `dict` with a key for each input variable and
                       a corresponding value that is an array of the input data. If specified as a plain list, then the
                       order is assumed the same as `Component.inputs`.
        :param model_fidelity: Fidelity indices to tune the model fidelity (model must request this
                               in its keyword arguments).
        :param output_path: Directory to save model output files (model must request this in its keyword arguments).
        :param executor: Executor for parallel execution if the model is not vectorized (optional).
        :param kwds: Additional keyword arguments to pass to the model (model must request these in its keyword args).
        :returns: The output data from the model, formatted as a `dict` with a key for each output variable and a
                  corresponding value that is an array of the output data.
        """
        # Format inputs to a common loop shape (fail if missing any)
        if len(inputs) == 0:
            return {}  # your fault
        if isinstance(inputs, list | np.ndarray):
            inputs = np.atleast_1d(inputs)
            inputs = {var.name: inputs[..., i] for i, var in enumerate(self.inputs)}

        var_shape = {}
        for var in self.inputs:
            s = None
            if (arr := kwds.get(f'{var.name}_coords')) is not None:
                s = arr.shape if len(arr.shape) == 1 else arr.shape[:-1]  # skip the coordinate dim (last axis)
            if var.compression is not None:
                for field in var.compression.fields:
                    var_shape[field] = s
            else:
                var_shape[var.name] = s
        inputs, loop_shape = format_inputs(inputs, var_shape=var_shape)

        N = int(np.prod(loop_shape))
        list_alpha = isinstance(model_fidelity, list | np.ndarray)
        alpha_requested = self.model_kwarg_requested('model_fidelity')
        for var in self.inputs:
            if var.compression is not None:
                for field in var.compression.fields:
                    if field not in inputs:
                        raise ValueError(f"Missing field '{field}' for input variable '{var}'.")
            elif var.name not in inputs:
                raise ValueError(f"Missing input variable '{var.name}'.")

        # Pass extra requested items to the model kwargs
        kwargs = copy.deepcopy(self.model_kwargs.data)
        if self.model_kwarg_requested('output_path'):
            kwargs['output_path'] = output_path
        if self.model_kwarg_requested('input_vars'):
            kwargs['input_vars'] = self.inputs
        if self.model_kwarg_requested('output_vars'):
            kwargs['output_vars'] = self.outputs
        if alpha_requested:
            if not list_alpha:
                model_fidelity = [model_fidelity] * N
            for i in range(N):
                if model_fidelity[i] == 'best':
                    model_fidelity[i] = self.max_alpha
                elif model_fidelity[i] == 'worst':
                    model_fidelity[i] = (0,) * len(self.model_fidelity)

        for k, v in kwds.items():
            if self.model_kwarg_requested(k):
                kwargs[k] = v

        # Compute model (vectorized, executor parallel, or serial)
        errors = {}
        if self.vectorized:
            if alpha_requested:
                kwargs['model_fidelity'] = np.atleast_1d(model_fidelity).reshape((N, -1))
            output_dict = self.model(*[inputs[var.name] for var in self.inputs], **kwargs) if self.call_unpacked \
                else self.model(inputs, **kwargs)
            if self.ret_unpacked:
                output_dict = (output_dict,) if not isinstance(output_dict, tuple) else output_dict
                output_dict = {out_var.name: output_dict[i] for i, out_var in enumerate(self.outputs)}
        else:
            if executor is None:  # Serial
                results = deque(maxlen=N)
                for i in range(N):
                    try:
                        if alpha_requested:
                            kwargs['model_fidelity'] = model_fidelity[i]
                        ret = self.model(*[{k: v[i] for k, v in inputs.items()}[var.name] for var in self.inputs],
                                         **kwargs) if self.call_unpacked else (
                            self.model({k: v[i] for k, v in inputs.items()}, **kwargs))
                        if self.ret_unpacked:
                            ret = (ret,) if not isinstance(ret, tuple) else ret
                            ret = {out_var.name: ret[i] for i, out_var in enumerate(self.outputs)}
                        results.append(ret)
                    except Exception:
                        results.append({'inputs': {k: v[i] for k, v in inputs.items()}, 'index': i,
                                        'model_kwargs': kwargs.copy(), 'error': traceback.format_exc()})
            else:  # Parallel
                results = deque(maxlen=N)
                futures = []
                for i in range(N):
                    if alpha_requested:
                        kwargs['model_fidelity'] = model_fidelity[i]
                    fs = executor.submit(self.model,
                                         *[{k: v[i] for k, v in inputs.items()}[var.name] for var in self.inputs],
                                         **kwargs) if self.call_unpacked else (
                        executor.submit(self.model, {k: v[i] for k, v in inputs.items()}, **kwargs))
                    futures.append(fs)
                wait(futures, timeout=None, return_when=ALL_COMPLETED)

                for i, fs in enumerate(futures):
                    try:
                        if alpha_requested:
                            kwargs['model_fidelity'] = model_fidelity[i]
                        ret = fs.result()
                        if self.ret_unpacked:
                            ret = (ret,) if not isinstance(ret, tuple) else ret
                            ret = {out_var.name: ret[i] for i, out_var in enumerate(self.outputs)}
                        results.append(ret)
                    except Exception:
                        results.append({'inputs': {k: v[i] for k, v in inputs.items()}, 'index': i,
                                        'model_kwargs': kwargs.copy(), 'error': traceback.format_exc()})

            # Collect parallel/serial results
            output_dict = {}
            for i in range(N):
                res = results.popleft()
                if 'error' in res:
                    errors[i] = res
                else:
                    for key, val in res.items():
                        # Save numeric outputs
                        numeric_flag = False
                        for var in self.outputs:
                            if var.compression is not None:  # field quantity return values
                                if key in var.compression.fields:
                                    if output_dict.get(key) is None:
                                        output_dict.setdefault(key, np.full((N, *np.atleast_1d(val).shape), np.nan))
                                    output_dict[key][i, ...] = np.atleast_1d(val)
                                    numeric_flag = True
                                    break
                            elif key == var:
                                if output_dict.get(key) is None:
                                    output_dict.setdefault(key, np.full((N, *np.atleast_1d(val).shape), np.nan))
                                output_dict[key][i, ...] = np.atleast_1d(val)
                                numeric_flag = True
                                break
                        # Otherwise, save other objects
                        if not numeric_flag:
                            if key == 'model_cost':
                                if output_dict.get(key) is None:
                                    output_dict.setdefault(key, np.full((N,), np.nan))
                                output_dict[key][i] = val
                            else:
                                if output_dict.get(key) is None:
                                    output_dict.setdefault(key, np.full((N,), None, dtype=object))
                                output_dict[key][i] = val

        # Save average model costs for each alpha fidelity
        if model_fidelity is not None and output_dict.get('model_cost') is not None:
            alpha_costs = {}
            for i, cost in enumerate(output_dict['model_cost']):
                alpha_costs.setdefault(MultiIndex(model_fidelity[i]), [])
                alpha_costs[MultiIndex(model_fidelity[i])].append(cost)
            for a, costs in alpha_costs.items():
                self.model_costs.setdefault(a, np.empty(0))
                self.model_costs[a] = np.nanmean(np.hstack((costs, self.model_costs[a])))

        # Reshape loop dimensions to match the original input shape
        output_dict = format_outputs(output_dict, loop_shape)

        for var in self.outputs:
            if var.compression is not None:
                for field in var.compression.fields:
                    if field not in output_dict:
                        self.logger.warning(f"Model return missing field '{field}' for output variable '{var}'. "
                                            f"Returning NaNs...")
                        output_dict[field].setdefault(field, np.full((N,), np.nan))
            elif var.name not in output_dict:
                self.logger.warning(f"Model return missing output variable '{var.name}'. Returning NaNs...")
                output_dict[var.name] = np.full((N,), np.nan)

        # Return the output dictionary and any errors
        if errors:
            output_dict['errors'] = errors
        return output_dict

    def predict(self, inputs: dict | Dataset,
                use_model: Literal['best', 'worst'] | tuple = None,
                model_dir: str | Path = None,
                index_set: Literal['train', 'test'] | IndexSet = 'test',
                misc_coeff: MiscTree = None,
                incremental: bool = False,
                executor: Executor = None,
                **kwds) -> Dataset:
        """Evaluate the MISC surrogate approximation at new inputs `x`.

        !!! Note "Using the underlying model"
            By default this will predict the MISC surrogate approximation; all inputs are assumed to be in a compressed
            and normalized form. If the component does not have a surrogate (i.e. it is analytical), then the inputs
            will be converted to model form and the underlying model will be called in place. If you instead want to
            override the surrogate, passing `use_model` will call the underlying model directly. In that case, the
            inputs should be passed in already in model form (i.e. full fields, denormalized).

        :param inputs: `dict` of input arrays for each variable input
        :param use_model: 'best'=high-fidelity, 'worst'=low-fidelity, tuple=a specific `alpha`, None=surrogate (default)
        :param model_dir: directory to save output files if `use_model` is specified, ignored otherwise
        :param index_set: the active index set, defaults to `self.active_set` if `'train'` or both
                          `self.active_set + self.candidate_set` if `'test'`
        :param misc_coeff: the data structure holding the MISC coefficients to use, which defaults to the
                           training or testing coefficients depending on the `index_set` parameter.
        :param incremental: a special flag to use if the provided `index_set` is an incremental update to the active
                            index set. A temporary copy of the internal `misc_coeff` data structure will be updated
                            and used to incorporate the new indices.
        :param executor: executor for parallel execution if the model is not vectorized (optional), will use the
                         executor for looping over MISC coefficients if evaluating the surrogate rather than the model
        :param kwds: additional keyword arguments to pass to the model (if using the underlying model)
        :returns: the surrogate approximation of the model (or the model return itself if `use_model`)
        """
        # Use raw model inputs/outputs
        if use_model is not None:
            outputs = self.call_model(inputs, model_fidelity=use_model, output_path=model_dir, executor=executor,**kwds)
            ret = {}
            for var in self.outputs:
                if var in outputs:
                    ret[var.name] = outputs[var.name]
                elif var.compression is not None:
                    for field in var.compression.fields:
                        ret[field] = outputs[field]
            return ret

        # Convert inputs/outputs to/from model if no surrogate (i.e. analytical models)
        if not self.has_surrogate:
            field_coords = {f'{var}_coords': self.model_kwargs.get(f'{var}_coords', kwds.get(f'{var}_coords', None))
                            for var in self.inputs}
            inputs, field_coords = to_model_dataset(inputs, self.inputs, del_latent=True, **field_coords)
            field_coords.update(kwds)
            outputs = self.call_model(inputs, model_fidelity=use_model or 'best', output_path=model_dir,
                                      executor=executor, **field_coords)
            outputs, surr_vars = to_surrogate_dataset(outputs, self.outputs, del_fields=True, **field_coords)
            return {str(var): outputs[var] for var in surr_vars}

        # Choose the correct index set and misc_coeff data structures
        if incremental:
            misc_coeff = copy.deepcopy(self.misc_coeff_train)
            self.update_misc_coeff(index_set, self.active_set, misc_coeff)
            index_set = self.active_set.union(index_set)
        else:
            index_set, misc_coeff = self._match_index_set(index_set, misc_coeff)

        # Format inputs for surrogate prediction (all scalars at this point, including latent coeffs)
        inputs, loop_shape = format_inputs(inputs)  # {'x': (N,)}
        outputs = {}

        # Handle prediction with empty active set (return nan)
        if len(index_set) == 0:
            for var in self.outputs:
                outputs[var.name] = np.full(loop_shape, np.nan)
            return outputs

        y_vars = self._surrogate_outputs()  # Only request this component's specified outputs (ignore all extras)

        # Combination technique MISC surrogate prediction
        results = []
        coeffs = []
        for alpha, beta in index_set:
            comb_coeff = misc_coeff[alpha, beta]
            if np.abs(comb_coeff) > 0:
                coeffs.append(comb_coeff)
                args = (self.misc_states.get((alpha, beta)),
                        self.training_data.get(alpha, beta[:len(self.data_fidelity)], skip_nan=True, y_vars=y_vars))

                results.append(self.interpolator.predict(inputs, *args) if executor is None else
                               executor.submit(self.interpolator.predict, inputs, *args))

        if executor is not None:
            wait(results, timeout=None, return_when=ALL_COMPLETED)
            results = [future.result() for future in results]

        for coeff, interp_pred in zip(coeffs, results):
            for var, arr in interp_pred.items():
                if outputs.get(var) is None:
                    outputs[str(var)] = coeff * arr
                else:
                    outputs[str(var)] += coeff * arr

        return format_outputs(outputs, loop_shape)

    def update_misc_coeff(self, new_indices: IndexSet, index_set: Literal['test', 'train'] | IndexSet = 'train',
                          misc_coeff: MiscTree = None):
        """Update MISC coefficients incrementally resulting from the addition of new indices to an index set.

        !!! Warning "Incremental updates"
            This function is used to update the MISC coefficients stored in `misc_coeff` after adding new indices
            to the given `index_set`. If a custom `index_set` or `misc_coeff` are provided, the user is responsible
            for ensuring the data structures are consistent. Since this is an incremental update, this means all
            existing coefficients for every index in `index_set` should be precomputed and stored in `misc_coeff`.

        :param new_indices: a set of $(\\alpha, \\beta)$ tuples that are being added to the `index_set`
        :param index_set: the active index set, defaults to `self.active_set` if `'train'` or both
                          `self.active_set + self.candidate_set` if `'test'`
        :param misc_coeff: the data structure holding the MISC coefficients to update, which defaults to the
                           training or testing coefficients depending on the `index_set` parameter. This data structure
                           is modified in place.
        """
        index_set, misc_coeff = self._match_index_set(index_set, misc_coeff)

        for new_alpha, new_beta in new_indices:
            new_ind = np.array(new_alpha + new_beta)

            # Update all existing/new coefficients if they are a distance of [0, 1] "below" the new index
            # Note that new indices can only be [0, 1] away from themselves -- not any other new indices
            for old_alpha, old_beta in itertools.chain(index_set, [(new_alpha, new_beta)]):
                old_ind = np.array(old_alpha + old_beta)
                diff = new_ind - old_ind
                if np.all(np.isin(diff, [0, 1])):
                    if misc_coeff.get((old_alpha, old_beta)) is None:
                        misc_coeff[old_alpha, old_beta] = 0
                    misc_coeff[old_alpha, old_beta] += (-1) ** int(np.sum(np.abs(diff)))

    def activate_index(self, alpha: MultiIndex, beta: MultiIndex, model_dir: str | Path = None,
                       executor: Executor = None):
        """Add a multi-index to the active set and all neighbors to the candidate set.

        !!! Warning
            The user of this function is responsible for ensuring that the index set maintains downward-closedness.
            That is, only activate indices that are neighbors of the current active set.

        :param alpha: A multi-index specifying model fidelity
        :param beta: A multi-index specifying surrogate fidelity
        :param model_dir: Directory to save model output files
        :param executor: Executor for parallel execution of model on training data if the model is not vectorized
        """
        if (alpha, beta) in self.active_set:
            self.logger.warning(f'Multi-index {(alpha, beta)} is already in the active index set. Ignoring...')
            return
        if (alpha, beta) not in self.candidate_set and (sum(alpha) + sum(beta)) > 0:
            # Can only activate the initial index (0, 0, ... 0) without it being in the candidate set
            self.logger.warning(f'Multi-index {(alpha, beta)} is not a neighbor of the active index set, so it '
                                f'cannot be activated. Please only add multi-indices from the candidate set. '
                                f'Ignoring...')
            return

        # Collect all neighbor candidate indices
        neighbors = self._neighbors(alpha, beta, forward=True)
        indices = list(itertools.chain([(alpha, beta)] if (alpha, beta) not in self.candidate_set else [], neighbors))

        # Refine and collect all new model inputs (i.e. training points) requested by the new candidates
        alpha_list = []    # keep track of model fidelities
        design_list = []   # keep track of training data coordinates/locations/indices
        model_inputs = {}  # concatenate all model inputs
        field_coords = {f'{var}_coords': self.model_kwargs.get(f'{var}_coords', None) for var in self.inputs}
        domains = self.inputs.get_domains()
        weight_fcns = self.inputs.get_pdfs()
        for a, b in indices:
            design_coords, design_pts = self.training_data.refine(a, b[:len(self.data_fidelity)],
                                                                  domains, weight_fcns)
            design_pts, fc = to_model_dataset(design_pts, self.inputs, del_latent=True, **field_coords)

            # Remove duplicate (alpha, coords) pairs -- so you don't evaluate the model twice for the same input
            i = 0
            del_idx = []
            for other_design in design_list:
                for other_coord in other_design:
                    for j, curr_coord in enumerate(design_coords):
                        if curr_coord == other_coord and a == alpha_list[i] and j not in del_idx:
                            del_idx.append(j)
                    i += 1
            design_coords = [design_coords[j] for j in range(len(design_coords)) if j not in del_idx]
            design_pts = {var: np.delete(arr, del_idx, axis=0) for var, arr in design_pts.items()}

            alpha_list.extend([tuple(a)] * len(design_coords))
            design_list.append(design_coords)
            field_coords.update(fc)
            for var in design_pts:
                model_inputs[var] = design_pts[var] if model_inputs.get(var) is None else (
                    np.concatenate((model_inputs[var], design_pts[var]), axis=0))

        # Evaluate model at designed training points
        if len(alpha_list) > 0:
            self.logger.info(f"Running {len(alpha_list)} total model evaluations for component "
                             f"'{self.name}' new candidate indices: {indices}...")
            model_outputs = self.call_model(model_inputs, model_fidelity=alpha_list, output_path=model_dir,
                                            executor=executor, **field_coords)
            errors = model_outputs.pop('errors', {})

        # Unpack model outputs and update states
        start_idx = 0
        for i, (a, b) in enumerate(indices):
            num_train_pts = len(design_list[i])
            end_idx = start_idx + num_train_pts  # Ensure loop dim of 1 gets its own axis (might have been squeezed)
            yi_dict = {var: arr[np.newaxis, ...] if len(alpha_list) == 1 and arr.shape[0] != 1 else
                       arr[start_idx:end_idx, ...] for var, arr in model_outputs.items()}

            # Check for errors and store
            err_coords = []
            err_list = []
            for idx in list(errors.keys()):
                if idx < end_idx:
                    err_info = errors.pop(idx)
                    err_info['index'] = idx - start_idx
                    err_coords.append(design_list[i][idx - start_idx])
                    err_list.append(err_info)
            if len(err_list) > 0:
                self.logger.warning(f"Model errors occurred while adding candidate ({a}, {b}) for component "
                                    f"{self.name}. Leaving NaN values in training data...")
                self.training_data.set_errors(a, b[:len(self.data_fidelity)], err_coords, err_list)

            # Compress field quantities and normalize
            yi_dict, y_vars = to_surrogate_dataset(yi_dict, self.outputs, del_fields=False, **field_coords)

            # Store training data, computational cost, and new interpolator state
            self.training_data.set(a, b[:len(self.data_fidelity)], design_list[i], yi_dict)
            self.training_data.impute_missing_data(a, b[:len(self.data_fidelity)])
            self.misc_costs[a, b] = self.model_costs.get(a, 1.) * num_train_pts
            self.misc_states[a, b] = self.interpolator.refine(b[len(self.data_fidelity):],
                                                              self.training_data.get(a, b[:len(self.data_fidelity)],
                                                                                     y_vars=y_vars, skip_nan=True),
                                                              self.misc_states.get((alpha, beta)),
                                                              domains)
            start_idx = end_idx

        # Move to the active index set
        s = set()
        s.add((alpha, beta))
        self.update_misc_coeff(IndexSet(s), index_set='train')
        if (alpha, beta) in self.candidate_set:
            self.candidate_set.remove((alpha, beta))
        else:
            # Only for initial index which didn't come from the candidate set
            self.update_misc_coeff(IndexSet(s), index_set='test')
        self.active_set.update(s)

        self.update_misc_coeff(neighbors, index_set='test')  # neighbors will only ever pass through here once
        self.candidate_set.update(neighbors)

    def gradient(self, inputs: dict | Dataset,
                 index_set: Literal['train', 'test'] | IndexSet = 'test',
                 misc_coeff: MiscTree = None,
                 derivative: Literal['first', 'second'] = 'first',
                 executor: Executor = None) -> Dataset:
        """Evaluate the Jacobian or Hessian of the MISC surrogate approximation at new `inputs`, i.e.
        the first or second derivatives, respectively.

        :param inputs: `dict` of input arrays for each variable input
        :param index_set: the active index set, defaults to `self.active_set` if `'train'` or both
                          `self.active_set + self.candidate_set` if `'test'`
        :param misc_coeff: the data structure holding the MISC coefficients to use, which defaults to the
                           training or testing coefficients depending on the `index_set` parameter.
        :param derivative: whether to compute the first or second derivative (i.e. Jacobian or Hessian)
        :param executor: executor for looping over MISC coefficients (optional)
        :returns: a `dict` of the Jacobian or Hessian of the surrogate approximation for each output variable
        """
        if not self.has_surrogate:
            self.logger.warning("No surrogate model available for gradient computation.")
            return None

        index_set, misc_coeff = self._match_index_set(index_set, misc_coeff)
        inputs, loop_shape = format_inputs(inputs)  # {'x': (N,)}
        outputs = {}

        if len(index_set) == 0:
            for var in self.outputs:
                outputs[var] = np.full(loop_shape, np.nan)
            return outputs
        y_vars = self._surrogate_outputs()

        # Combination technique MISC gradient prediction
        results = []
        coeffs = []
        for alpha, beta in index_set:
            comb_coeff = misc_coeff[alpha, beta]
            if np.abs(comb_coeff) > 0:
                coeffs.append(comb_coeff)
                func = self.interpolator.gradient if derivative == 'first' else self.interpolator.hessian
                args = (self.misc_states.get((alpha, beta)),
                        self.training_data.get(alpha, beta[:len(self.data_fidelity)], skip_nan=True, y_vars=y_vars))

                results.append(func(inputs, *args) if executor is None else executor.submit(func, inputs, *args))

        if executor is not None:
            wait(results, timeout=None, return_when=ALL_COMPLETED)
            results = [future.result() for future in results]

        for coeff, interp_pred in zip(coeffs, results):
            for var, arr in interp_pred.items():
                if outputs.get(var) is None:
                    outputs[str(var)] = coeff * arr
                else:
                    outputs[str(var)] += coeff * arr

        return format_outputs(outputs, loop_shape)

    def hessian(self, *args, **kwargs):
        """Alias for `Component.gradient(*args, derivative='second', **kwargs)`."""
        return self.gradient(*args, derivative='second', **kwargs)

    def model_kwarg_requested(self, kwarg_name: str) -> bool:
        """Return whether the underlying component model requested this `kwarg_name`. Special kwargs include:

        - `output_path` — a save directory created by `amisc` will be passed to the model for saving model output files.
        - `alpha` — a tuple or list of model fidelity indices will be passed to the model to adjust fidelity.
        - `input_vars` — a list of `Variable` objects will be passed to the model for input variable information.
        - `output_vars` — a list of `Variable` objects will be passed to the model for output variable information.

        :param kwarg_name: the argument to check for in the underlying component model's function signature kwargs
        :returns: whether the component model requests this `kwarg` argument
        """
        signature = inspect.signature(self.model)
        for param in signature.parameters.values():
            if param.name == kwarg_name and param.default != param.empty:
                return True
        return False

    def set_logger(self, log_file: str | Path = None, stdout: bool = None, logger: logging.Logger = None,
                   level: int = logging.INFO):
        """Set a new `logging.Logger` object.

        :param log_file: log to file (if provided)
        :param stdout: whether to connect the logger to console (defaults to whatever is currently set or False)
        :param logger: the logging object to use (if None, then a new logger is created; this will override
                       the `log_file` and `stdout` arguments if set)
        :param level: the logging level to set (default is `logging.INFO`)
        """
        if stdout is None:
            stdout = False
            if self._logger is not None:
                for handler in self._logger.handlers:
                    if isinstance(handler, logging.StreamHandler):
                        stdout = True
                        break
        self._logger = logger or get_logger(self.name, log_file=log_file, stdout=stdout, level=level)

    def update_model(self, new_model: callable = None, model_kwargs: dict = None, **kwargs):
        """Update the underlying component model or its kwargs."""
        if new_model is not None:
            self.model = new_model
        new_kwargs = self.model_kwargs.data
        new_kwargs.update(model_kwargs or {})
        new_kwargs.update(kwargs)
        self.model_kwargs = new_kwargs

    def get_cost(self, alpha: MultiIndex, beta: MultiIndex) -> float:
        """Return the total cost (wall time s) required to add $(\\alpha, \\beta)$ to the MISC approximation.

        :param alpha: A multi-index specifying model fidelity
        :param beta: A multi-index specifying surrogate fidelity
        :returns: the total cost of adding this multi-index pair to the MISC approximation
        """
        try:
            return self.misc_costs[alpha, beta]
        except Exception:
            return 0.0

    @staticmethod
    def is_downward_closed(indices: IndexSet) -> bool:
        """Return if a list of $(\\alpha, \\beta)$ multi-indices is downward-closed.

        MISC approximations require a downward-closed set in order to use the combination-technique formula for the
        coefficients (as implemented by `Component.update_misc_coeff()`).

        !!! Example
            The list `[( (0,), (0,) ), ( (1,), (0,) ), ( (1,), (1,) )]` is downward-closed. You can visualize this as
            building a stack of cubes: in order to place a cube, all adjacent cubes must be present (does the logo
            make sense now?).

        :param indices: `IndexSet` of (`alpha`, `beta`) multi-indices
        :returns: whether the set of indices is downward-closed
        """
        # Iterate over every multi-index
        for alpha, beta in indices:
            # Every smaller multi-index must also be included in the indices list
            sub_sets = [np.arange(tuple(alpha + beta)[i] + 1) for i in range(len(alpha) + len(beta))]
            for ele in itertools.product(*sub_sets):
                tup = (MultiIndex(ele[:len(alpha)]), MultiIndex(ele[len(alpha):]))
                if tup not in indices:
                    return False
        return True

    def clear(self):
        """Clear the component of all training data, index sets, and MISC states."""
        self.active_set.clear()
        self.candidate_set.clear()
        self.misc_states.clear()
        self.misc_costs.clear()
        self.misc_coeff_train.clear()
        self.misc_coeff_test.clear()
        self.model_costs.clear()
        self.training_data.clear()

    def serialize(self, keep_yaml_objects: bool = False, serialize_args: dict[str, tuple] = None,
                  serialize_kwargs: dict[str: dict] = None) -> dict:
        """Convert to a `dict` with only standard Python types as fields and values.

        :param keep_yaml_objects: whether to keep `Variable` or other yaml serializable objects instead of
                                  also serializing them (default is False)
        :param serialize_args: additional arguments to pass to the `serialize` method of each `Component` attribute;
                               specify as a `dict` of attribute names to tuple of arguments to pass
        :param serialize_kwargs: additional keyword arguments to pass to the `serialize` method of each
                                 `Component` attribute
        :returns: a `dict` representation of the `Component` object
        """
        serialize_args = serialize_args or dict()
        serialize_kwargs = serialize_kwargs or dict()
        d = {}
        for key, value in self.__dict__.items():
            if value is not None and not key.startswith('_'):
                if key == 'serializers':
                    # Update the serializers
                    serializers = self._validate_serializers({k: type(getattr(self, k)) for k in value.keys()})
                    d[key] = {k: (v.obj if keep_yaml_objects else v.serialize()) for k, v in serializers.items()}
                elif key in ['inputs', 'outputs'] and not keep_yaml_objects:
                    d[key] = value.serialize(**serialize_kwargs.get(key, {}))
                elif key == 'model' and not keep_yaml_objects:
                    d[key] = YamlSerializable(obj=value).serialize()
                elif key in ['data_fidelity', 'surrogate_fidelity', 'model_fidelity']:
                    if len(value) > 0:
                        d[key] = str(value)
                elif key in ['active_set', 'candidate_set']:
                    if len(value) > 0:
                        d[key] = value.serialize()
                elif key in ['misc_costs', 'misc_coeff_train', 'misc_coeff_test', 'misc_states']:
                    if len(value) > 0:
                        d[key] = value.serialize(keep_yaml_objects=keep_yaml_objects)
                elif key in ['model_costs']:
                    if len(value) > 0:
                        d[key] = {str(k): float(v) for k, v in value.items()}
                elif key in ComponentSerializers.__annotations__.keys():
                    if key in ['training_data', 'interpolator'] and not self.has_surrogate:
                        continue
                    else:
                        d[key] = value.serialize(*serialize_args.get(key, ()), **serialize_kwargs.get(key, {}))
                else:
                    d[key] = value
        return d

    @classmethod
    def deserialize(cls, serialized_data: dict, search_paths: list[str | Path] = None,
                    search_keys: list[str] = None) -> Component:
        """Return a `Component` from `data`. Let pydantic handle field validation and conversion. If any component
        data has been saved to file and the save file doesn't exist, then the loader will search for the file
        in the current working directory and any additional search paths provided.

        :param serialized_data: the serialized data to construct the object from
        :param search_paths: paths to try and find any save files (i.e. if they moved since they were serialized),
                             will always search in the current working directory by default
        :param search_keys: keys to search for save files in each component (default is all keys in
                            [`ComponentSerializers`][amisc.component.ComponentSerializers], in addition to variable
                            inputs and outputs)
        """
        if isinstance(serialized_data, Component):
            return serialized_data
        elif callable(serialized_data):
            # try to construct a component from a raw function (assume data fidelity is (2,) for each inspected input)
            return cls(serialized_data, data_fidelity=(2,) * len(_inspect_function(serialized_data)[0]))

        search_paths = search_paths or []
        search_keys = search_keys or []
        search_keys.extend(ComponentSerializers.__annotations__.keys())
        comp = serialized_data

        for key in search_keys:
            if (filename := comp.get(key, None)) is not None:
                comp[key] = search_for_file(filename, search_paths=search_paths)  # will ret original str if not found

        for key in ['inputs', 'outputs']:
            for var in comp.get(key, []):
                if isinstance(var, dict):
                    if (compression := var.get('compression', None)) is not None:
                        var['compression'] = search_for_file(compression, search_paths=search_paths)

        return cls(**comp)

    @staticmethod
    def _yaml_representer(dumper: yaml.Dumper, comp: Component) -> yaml.MappingNode:
        """Convert a single `Component` object (`data`) to a yaml MappingNode (i.e. a `dict`)."""
        save_path, save_file = _get_yaml_path(dumper)
        serialize_kwargs = {}
        for key, serializer in comp.serializers.items():
            if issubclass(serializer.obj, PickleSerializable):
                filename = save_path / f'{save_file}_{comp.name}_{key}.pkl'
                serialize_kwargs[key] = {'save_path': save_path / filename}
        return dumper.represent_mapping(Component.yaml_tag, comp.serialize(serialize_kwargs=serialize_kwargs,
                                                                           keep_yaml_objects=True))

    @staticmethod
    def _yaml_constructor(loader: yaml.Loader, node):
        """Convert the `!Component` tag in yaml to a `Component` object."""
        # Add a file search path in the same directory as the yaml file being loaded from
        save_path, save_file = _get_yaml_path(loader)
        if isinstance(node, yaml.SequenceNode):
            return [ele if isinstance(ele, Component) else Component.deserialize(ele, search_paths=[save_path])
                    for ele in loader.construct_sequence(node, deep=True)]
        elif isinstance(node, yaml.MappingNode):
            return Component.deserialize(loader.construct_mapping(node, deep=True), search_paths=[save_path])
        else:
            raise NotImplementedError(f'The "{Component.yaml_tag}" yaml tag can only be used on a yaml sequence or '
                                      f'mapping, not a "{type(node)}".')



"""Module for data transformation methods.

Includes:

- `Transform` — an abstract interface for specifying a transformation.
- `Linear` — a linear transformation $y=mx+b$.
- `Log` — a logarithmic transformation $y=\\log_b(x + \\mathrm{offset})$.
- `Minmax` — a min-max scaling transformation $x: (lb, ub) \\mapsto (lb_{norm}, ub_{norm})$.
- `Zscore` — a z-score normalization transformation $y=(x-\\mu)/\\sigma$.

Transform objects can be converted easily to/from strings for serialization.
"""
from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
from numpy.typing import ArrayLike

from amisc.utils import parse_function_string

__all__ = ['Transform', 'Linear', 'Log', 'Minmax', 'Zscore']


class Transform(ABC):
    """A base class for all transformations.

    :ivar transform_args: the arguments for the transformation
    :vartype transform_args: tuple
    """

    def __init__(self, transform_args: tuple):
        self.transform_args = transform_args

    def __str__(self):
        """Serialize a `Transform` object to string."""
        return f'{type(self).__name__}{self.transform_args}'

    @classmethod
    def from_string(cls, transform_spec: str | list[str]) -> list[Transform] | None:
        """Return a list of `Transforms` given a list of string specifications. Available transformations are:

        - **linear** — $x_{norm} = mx + b$ specified as `linear(m, b)` or `linear(slope=m, offset=b)`. `m=1, b=0` if not
                       specified.
        - **log** — $x_{norm} = \\log_b(x)$ specified as `log` or `log10` for the natural or common logarithms. For a
                    different base, use `log(b)`. Optionally, specify `offset` for `log(x+offset)`.
        - **minmax** — $x_{norm} = \\frac{x - a}{b - a}(u - l) + l$ specified as `minmax(a, b, l, u)` or
                       `minmax(lb=a, ub=b, lb_norm=l, ub_norm=u)`. Scales `x` from the range `(a, b)` to `(l, u)`. By
                       default, `(a, b)` is the Variable's domain and `(l, u)` is `(0, 1)`. Use simply as `minmax`
                       to use all defaults.
        - **zscore** — $x_{norm} = \\frac{x - m}{s}$ specified as `zscore(m, s)` or `zscore(mu=m, std=s)`. If the
                       Variable is specified as `distribution=normal`, then `zscore` defaults to the Variable's
                       own `mu, std`.

        !!! Example
            ```python
            transforms = Transform.from_string(['log10', 'linear(2, 4)'])
            print(transforms)
            ```
            will give
            ```shell
            [ Log(10), Linear(2, 4) ]  # the corresponding `Transform` objects
            ```

        !!! Warning
            You may optionally leave the `minmax` arguments blank to defer to the bounds of the parent `Variable`.
            You may also optionally leave the `zscore` arguments blank to defer to the `(mu, std)` of the parent
            `Variable`, but this will throw a runtime error if `Variable.distribution` is not `Normal(mu, std)`.
        """
        if transform_spec is None:
            return None
        if isinstance(transform_spec, str | Transform):
            transform_spec = [transform_spec]

        transforms = []
        for spec_string in transform_spec:
            if isinstance(spec_string, Transform):
                transforms.append(spec_string)
                continue

            name, args, kwargs = parse_function_string(spec_string)
            if name.lower() == 'linear':
                try:
                    slope = float(kwargs.get('slope', args[0] if len(args) > 0 else 1))
                    offset = float(kwargs.get('offset', args[1] if len(args) > 1 else 0))
                    transforms.append(Linear((slope, offset)))
                except Exception as e:
                    raise ValueError(f'Linear transform spec "{spec_string}" is not valid: Try "linear(m, b)".') from e
            elif name.lower() in ['log', 'log10']:
                try:
                    log_base = float(kwargs.get('base', args[0] if len(args) > 0 else (np.e if name.lower() == 'log'
                                                                                       else 10)))
                    offset = float(kwargs.get('offset', args[1] if len(args) > 1 else 0))
                    transforms.append(Log((log_base, offset)))
                except Exception as e:
                    raise ValueError(f'Log transform spec "{spec_string}" is not valid: Try "log(base, offset)"') from e
            elif name.lower() in ['minmax', 'maxabs']:
                try:
                    # Defer bounds to the Variable by setting np.nan
                    lb = float(kwargs.get('lb', args[0] if len(args) > 0 else np.nan))
                    ub = float(kwargs.get('ub', args[1] if len(args) > 1 else np.nan))
                    lb_norm = float(kwargs.get('lb_norm', args[2] if len(args) > 2 else 0))
                    ub_norm = float(kwargs.get('ub_norm', args[3] if len(args) > 3 else 1))
                    transforms.append(Minmax((lb, ub, lb_norm, ub_norm)))
                except Exception as e:
                    raise ValueError(f'Minmax transform spec "{spec_string}" is not valid: Try "minmax(lb, ub)"') from e
            elif name.lower() in ['z', 'zscore']:
                try:
                    # Defer (mu, std) to the Variable by setting np.nan
                    mu = float(kwargs.get('mu', args[0] if len(args) > 0 else np.nan))
                    std = float(kwargs.get('std', args[1] if len(args) > 1 else np.nan))
                    transforms.append(Zscore((mu, std)))
                except Exception as e:
                    raise ValueError(f'Z-score normalization string "{spec_string}" is not valid: '
                                     f'Try "zscore(mu, std)".') from e
            else:
                raise NotImplementedError(f'Transform method "{name}" is not implemented.')

        return transforms

    def transform(self, x: ArrayLike, inverse: bool = False, transform_args: tuple = None) -> ArrayLike:
        """Transform the given values `x`. This wrapper function handles the input type and tries to
         return the transformed values in the same type.

        :param x: the values to transform
        :param inverse: whether to do the inverse transform instead
        :param transform_args: overrides `Transform.transform_args`
        :return: the transformed values
        """
        input_type = type(x)
        result = self._transform(np.atleast_1d(x), inverse, transform_args)
        if input_type in [int, float]:
            return float(result[0])
        elif input_type is list:
            return result.tolist()
        elif input_type is tuple:
            return tuple(result.tolist())
        else:
            return result  # just keep as np.ndarray for everything else

    @abstractmethod
    def _transform(self, x, inverse=False, transform_args=None):
        """Abstract method that subclass `Transform` objects should implement.

        :param x: the values to transform
        :param inverse: whether to do the inverse transform instead
        :param transform_args: overrides `Transform.transform_args`
        :return: the transformed values
        """
        raise NotImplementedError


class Linear(Transform):
    """A Linear transform: $y=mx+b$.

    :ivar transform_args: `(m, b)` the slope and offset
    """
    def _transform(self, x, inverse=False, transform_args=None):
        slope, offset = transform_args or self.transform_args
        return (x - offset) / slope if inverse else slope * x + offset


class Log(Transform):
    """A Log transform: $y=\\log_b(x + \\mathrm{offset})$.

    :ivar transform_args: `(base, offset)` the log base and offset
    """
    def _transform(self, x, inverse=False, transform_args=None):
        log_base, offset = transform_args or self.transform_args
        return np.exp(x * np.log(log_base)) - offset if inverse else np.log(x + offset) / np.log(log_base)


class Minmax(Transform):
    """A Minmax transform: $x: (lb, ub) \\mapsto (lb_{norm}, ub_{norm})$.

    :ivar transform_args: `(lb, ub, lb_norm, ub_norm)` the original lower and upper bounds and the normalized bounds
    """
    def _transform(self, x, inverse=False, transform_args=None):
        transform_args = transform_args or self.transform_args
        if np.any(np.isnan(transform_args)):
            raise RuntimeError(f'Transform args may have missing values: {transform_args}')
        lb, ub, lb_norm, ub_norm = transform_args
        if inverse:
            return (x - lb_norm) / (ub_norm - lb_norm) * (ub - lb) + lb
        else:
            return (x - lb) / (ub - lb) * (ub_norm - lb_norm) + lb_norm

    def update(self, lb=None, ub=None, lb_norm=None, ub_norm=None):
        """Update the parameters of this transform.

        :param lb: the lower bound in the original variable space
        :param ub: the upper bound in the original variable space
        :param lb_norm: the lower bound of the transformed space
        :param ub_norm: the upper bound of the transformed space
        """
        transform_args = (lb, ub, lb_norm, ub_norm)
        self.transform_args = tuple([ele if ele is not None else self.transform_args[i]
                                     for i, ele in enumerate(transform_args)])


class Zscore(Transform):
    """A Zscore transform: $y=(x-\\mu)/\\sigma$.

    :ivar transform_args: `(mu, std)` the mean and standard deviation
    """
    def _transform(self, x, inverse=False, transform_args=None):
        transform_args = transform_args or self.transform_args
        if np.any(np.isnan(transform_args)):
            raise RuntimeError(f'Transform args may have missing values: {transform_args}')
        mu, std = transform_args
        return x * std + mu if inverse else (x - mu) / std

    def update(self, mu=None, std=None):
        """Update the parameters of this transform.

        :param mu: the mean of the transform
        :param std: the standard deviation of the transform
        """
        transform_args = (mu, std)
        self.transform_args = tuple([ele if ele is not None else self.transform_args[i]
                                     for i, ele in enumerate(transform_args)])



"""Examples to get started."""
# ruff: noqa: F841

def simple():
    # --8<-- [start:simple]
    import numpy as np

    from amisc import System

    def fun1(x):
        y = x * np.sin(np.pi * x)
        return y

    def fun2(y):
        z = 1 / (1 + 25 * y ** 2)
        return z

    system = System(fun1, fun2)
    system.inputs()['x'].update_domain((0, 1))
    system.outputs()['y'].update_domain((0, 1))

    system.fit()

    x_test = system.sample_inputs(10)
    y_test = system.predict(x_test)
    # --8<-- [end:simple]


def single_component():
    # --8<-- [start:single]
    from amisc import Component, System, Variable

    def fun(inputs):
        return {'y': inputs['x'] ** 2}

    x = Variable(distribution='U(-1, 1)')
    y = Variable()
    component = Component(fun, x, y, data_fidelity=(2,))
    system = System(component)

    system.fit()
    system.predict({'x': 0.5})  # {y: 0.25}
    # --8<-- [end:single]


def fire_sat():
    # --8<-- [start:fire_sat]
    from amisc.examples.models import fire_sat_system

    system = fire_sat_system()

    xtest = system.sample_inputs(100, use_pdf=True)     # --> (100, xdim)
    ytest = system.predict(xtest, use_model='best')     # --> (100, ydim)

    system.fit(test_set=(xtest, ytest), estimate_bounds=True)

    print(f'Inputs: {system.inputs()}')
    print(f'Outputs: {system.outputs()}')

    # Plots
    input_vars = ['H', 'Po']
    output_vars = ['Vsat', 'Asa']
    system.plot_allocation()
    system.plot_slice(input_vars, output_vars, show_model=['best', 'worst'], random_walk=True)
    # --8<-- [end:fire_sat]


def field_quantity():
    # --8<-- [start:field_qty]
    import numpy as np

    from amisc import Component, System, Variable, to_model_dataset
    from amisc.compression import SVD

    def my_model(inputs):
        """Compute a field quantity as a function of `x`."""
        field = np.sin(inputs['x'] * np.linspace(0, 1, 100))
        return {'f': field}

    dof = 100         # Number of degrees of freedom (i.e. size of the field qty)
    num_samples = 50  # Number of samples
    data_matrix = np.random.rand(dof, num_samples)
    field_coords = np.linspace(0, 1, dof)

    compression = SVD(rank=4, coords=field_coords, data_matrix=data_matrix)

    scalar = Variable('x', domain=(0, 1))
    field_qty = Variable('f', compression=compression)

    model = Component(my_model, inputs=scalar, outputs=field_qty, data_fidelity=(2,))
    system = System(model)

    system.fit(max_iter=10)

    xtest = system.sample_inputs(1000)
    ytest = system.predict(xtest)  # Will estimate the field qty in the latent/compressed space

    # Reconstruct the full field quantity
    ytest_reconstructed, _ = to_model_dataset(ytest, system.outputs())
    # --8<-- [end:field_qty]


if __name__ == '__main__':
    simple()
    single_component()
    fire_sat()
    field_quantity()



"""Provides some example test functions for demonstration purposes."""
import pickle
import uuid
from pathlib import Path

import numpy as np

from amisc.system import System


def f1(x):
    """Figure 5 in Jakeman 2022."""
    y1 = x * np.sin(np.pi * x)
    return y1


def f2(y1):
    """Figure 5 in Jakeman 2022."""
    y2 = 1 / (1 + 25*y1**2)
    return y2


def f3(x, y2):
    """Hypothetical third function for Figure 5 in Jakeman 2022."""
    y3 = x * np.cos(np.pi * y2)
    return y3


def tanh_func(inputs, A=2, L=1, frac=4):
    """Simple tunable tanh function."""
    return {'y': A * np.tanh(2 / (L/frac) * (inputs['x'] - L/2)) + A + 1}


def borehole_func(inputs):
    """Model found at https://www.sfu.ca/~ssurjano/borehole.html

    :param inputs: `dict` of input variables `rw`, `r`, `Tu`, `Hu`, `Tl`, `Hl`, `L`, and `Kw`
    :returns vdot: Water flow rate in m^3/yr
    """
    rw = inputs['rw']   # Radius of borehole (m)
    r = inputs['r']     # Radius of influence (m)
    Tu = inputs['Tu']   # Transmissivity (m^2/yr)
    Hu = inputs['Hu']   # Potentiometric head (m)
    Tl = inputs['Tl']   # Transmissivity (m^2/yr)
    Hl = inputs['Hl']   # Potentiometric head (m)
    L = inputs['L']     # Length of borehole (m)
    Kw = inputs['Kw']   # Hydraulic conductivity (m/yr)
    vdot = 2*np.pi*Tu*(Hu-Hl) / (np.log(r/rw) * (1 + (2*L*Tu/(np.log(r/rw)*Kw*rw**2)) + (Tu/Tl)))

    return {'vdot': vdot}


def wing_weight_func(inputs):
    """Model found at https://www.sfu.ca/~ssurjano/wingweight.html

    :param inputs: `dict` of input variables `Sw`, `Wfw`, `A`, `Lambda`, `q`, `lamb`, `tc`, `Nz`, `Wdg`, and `Wp`
    :returns Wwing: the weight of the airplane wing (lb)
    """
    Sw = inputs['Sw']           # Wing area (ft^2)
    Wfw = inputs['Wfw']         # Weight of fuel (lb)
    A = inputs['A']             # Aspect ratio
    Lambda = inputs['Lambda']   # Quarter-chord sweep (deg)
    q = inputs['q']             # Dynamic pressure (lb/ft^2)
    lamb = inputs['lambda']     # taper ratio
    tc = inputs['tc']           # Aerofoil thickness to chord ratio
    Nz = inputs['Nz']           # Ultimate load factor
    Wdg = inputs['Wdg']         # Design gross weight (lb)
    Wp = inputs['Wp']           # Paint weight (lb/ft^2)
    Lambda = Lambda*(np.pi/180)
    Wwing = 0.036*(Sw**0.758)*(Wfw**0.0035)*((A/(np.cos(Lambda))**2)**0.6)*(q**0.006)*(lamb**0.04)*\
            (100*tc/np.cos(Lambda))**(-0.3)*((Nz*Wdg)**0.49) + Sw*Wp

    return {'Wwing': Wwing}


def nonlinear_wave(inputs, env_var=0.1**2, wavelength=0.5, wave_amp=0.1, tanh_amp=0.5, L=1, t=0.25):
    """Custom nonlinear model of a traveling Gaussian wave for testing.

    :param inputs: `dict` of input variables `d` and `theta`
    :param env_var: variance of Gaussian envelope
    :param wavelength: sinusoidal perturbation wavelength
    :param wave_amp: amplitude of perturbation
    :param tanh_amp: amplitude of tanh(x)
    :param L: domain length of underlying tanh function
    :param t: transition length of tanh function (as fraction of L)
    :returns: `dict` with model output `y`
    """
    d = inputs['d']
    theta = inputs['theta']

    # Traveling sinusoid with moving Gaussian envelope
    env_range = [0.2, 0.6]
    mu = env_range[0] + theta * (env_range[1] - env_range[0])
    theta_env = 1 / (np.sqrt(2 * np.pi * env_var)) * np.exp(-0.5 * (d - mu) ** 2 / env_var)
    ftheta = wave_amp * np.sin((2*np.pi/wavelength) * theta) * theta_env

    # Underlying tanh dependence on d
    fd = tanh_amp * np.tanh(2/(L*t)*(d - L/2)) + tanh_amp

    # Compute model = f(theta, d) + f(d)
    return {'y': ftheta + fd}


def fire_sat_globals():
    """Global variables for the fire satellite system."""
    Re = 6378140.    # Radius of Earth (m)
    mu = 3.986e14    # Gravitational parameter (m^3 s^-2)
    eta = 0.22       # Power efficiency
    Id = 0.77        # Inherent degradation of the array
    thetai = 0.      # Sun incidence angle
    LT = 15.         # Spacecraft lifetime (years)
    eps = 0.0375     # Power production degradation (%/year)
    rlw = 3.         # Length to width ratio
    nsa = 3.         # Number of solar arrays
    rho = 700.       # Mass density of arrays (kg/m^3)
    t = 0.005        # Thickness (m)
    D = 2.           # Distance between panels (m)
    IbodyX = 6200.   # kg*m^2
    IbodyY = 6200.   # kg*m^2
    IbodyZ = 4700.   # kg*m^2
    dt_slew = 760.   # s
    theta = 15.      # Deviation of moment axis from vertical (deg)
    As = 13.85       # Area reflecting radiation (m^2)
    c = 2.9979e8     # Speed of light (m/s)
    M = 7.96e15      # Magnetic moment of earth (A*m^2)
    Rd = 5.          # Residual dipole of spacecraft (A*m^2)
    rhoa=5.148e-11   # Atmospheric density (kg/m^3) -- typo in Chaudhuri 2018 has this as 1e11 instead
    A = 13.85        # Cross-section in flight (m^2)
    Phold = 20.      # Holding power (W)
    omega = 6000.    # Max vel of wheel (rpm)
    nrw = 3.         # Number of reaction wheels

    return {k: v for k, v in locals().items() if not k.startswith('_') and isinstance(v, float)}


def orbit_fun(inputs, output_path=None, pct_failure=0):
    """Compute the orbit model for the fire satellite system.

    :param inputs: `dict` of input variables `H` and `Φ`
    :param output_path: where to save model outputs
    :param pct_failure: probability of a failure
    :returns: `dict` with model outputs `Vsat`, `To`, `Te`, and `Slew`
    """
    v = fire_sat_globals()                              # Global variables

    H = inputs['H']                                     # Altitude (m)
    phi = inputs[u'Φ']                                  # Target diameter (m)
    vel = np.sqrt(v['mu'] / (v['Re'] + H))              # Satellite velocity (m/s)
    dt_orbit = 2*np.pi*(v['Re'] + H) / vel                                                     # Orbit period (s)
    dt_eclipse = (dt_orbit/np.pi)*np.arcsin(v['Re'] / (v['Re'] + H))                           # Eclipse period (s)
    theta_slew = np.arctan(np.sin(phi / v['Re']) / (1 - np.cos(phi / v['Re']) + H/v['Re']))    # Max slew angle (rad)

    num_samples = np.atleast_1d(H).shape[0]

    if np.random.rand() < pct_failure:
        i = np.random.randint(0, num_samples)
        i2 = np.random.randint(0, num_samples)
        vel[i] = np.nan
        theta_slew[i2] = np.nan

    y = {'Vsat': vel, 'To': dt_orbit, 'Te': dt_eclipse, 'Slew': theta_slew}

    if output_path is not None:
        files = []
        id = str(uuid.uuid4())
        for index in range(num_samples):
            fname = f'{id}_{index}.pkl'
            with open(Path(output_path) / fname, 'wb') as fd:
                pickle.dump({var: np.atleast_1d(y[var])[index] for var in y.keys()}, fd)
            files.append(fname)
        y['output_path'] = files

    return y


def power_fun(inputs, model_fidelity=(0,), *, output_path=None, pct_failure=0):
    """Compute the power model for the fire satellite system.

    :param inputs: `dict` of input variables `Po`, `Fs`, `To`, `Te`, and `Pat`
    :param model_fidelity: fidelity index for the model
    :param output_path: where to save model outputs
    :param pct_failure: probability of a failure
    :returns: `dict` with model outputs `Imin`, `Imax`, `Ptot`, and `Asa`
    """
    alpha = model_fidelity
    pct = 1 - (2 - alpha[0]) * 0.04 if len(alpha) == 1 else 1  # extra pct error term
    v = fire_sat_globals()       # Global variables

    Po = inputs['Po']            # Other power sources (W)
    Fs = inputs['Fs']            # Solar flux (W/m^2)
    dt_orbit = inputs['To']      # Orbit period (s)
    dt_eclipse = inputs['Te']    # Eclipse period (s)
    Pacs = inputs['Pat']         # Power from attitude control system (W)

    Ptot = Po + Pacs
    Pe = Ptot
    Pd = Ptot
    Xe = 0.6                     # These are power efficiencies in eclipse and daylight
    Xd = 0.8                     # See Ch. 11 of Wertz 1999 SMAD
    Te = dt_eclipse
    Td = dt_orbit - Te
    Psa = ((Pe*Te/Xe) + (Pd*Td/Xd)) / Td
    Pbol = v['eta'] * Fs * v['Id'] * np.cos(v['thetai'])
    Peol = Pbol * (1 - v['eps']) ** v['LT']
    Asa = Psa / Peol                                # Total solar array size (m^2)
    L = np.sqrt(Asa * v['rlw'] / v['nsa'])
    W = np.sqrt(Asa / (v['rlw'] * v['nsa']))
    msa = 2 * v['rho'] * L * W * v['t']              # Mass of solar array
    Ix = msa*((1/12)*(L**2 + v['t']**2) + (v['D']+L/2)**2) + v['IbodyX']
    Iy = (msa/12)*(L**2 + W**2) + v['IbodyY']        # typo in Zaman 2013 has this as H**2 instead of L**2
    Iz = msa*((1/12)*(L**2 + W**2) + (v['D'] + L/2)**2) + v['IbodyZ']
    Itot = np.concatenate((Ix[..., np.newaxis], Iy[..., np.newaxis], Iz[..., np.newaxis]), axis=-1)
    Imin = np.min(Itot, axis=-1)
    Imax = np.max(Itot, axis=-1)

    num_samples = np.atleast_1d(Po).shape[0]

    if np.random.rand() < pct_failure:
        i = np.random.randint(0, num_samples)
        i2 = np.random.randint(0, num_samples)
        Imin[i2] = np.nan
        Asa[i] = np.nan

    y = {'Imin': Imin, 'Imax': Imax*pct, 'Ptot': Ptot*pct, 'Asa': Asa}

    if output_path is not None:
        files = []
        id = str(uuid.uuid4())
        for index in range(num_samples):
            fname = f'{id}_{index}.pkl'
            with open(Path(output_path) / fname, 'wb') as fd:
                pickle.dump({var: np.atleast_1d(y[var])[index] for var in y.keys()}, fd)
            files.append(fname)
        y['output_path'] = files

    return y


def attitude_fun(inputs, model_fidelity=(0,)):
    """Compute the attitude model for the fire satellite system.

    :param inputs: `dict` of input variables `H`, `Fs`, `Lsp`, `q`, `La`, `Cd`, `Vsat`, and `Slew`
    :param model_fidelity: fidelity index for the model
    :returns: `dict` with model outputs `Pat` and `tau_tot`
    """
    alpha = model_fidelity
    v = fire_sat_globals()       # Global variables
    pct = 1 - (2 - alpha[0])*0.04 if len(alpha) == 1 else 1  # extra model fidelity pct error term

    H = inputs['H']              # Altitude (m)
    Fs = inputs['Fs']            # Solar flux
    Lsp = inputs['Lsp']          # Moment arm for solar radiation pressure
    q = inputs['q']              # Reflectance factor
    La = inputs['La']            # Moment arm for aerodynamic drag
    Cd = inputs['Cd']            # Drag coefficient
    vel = inputs['Vsat']         # Satellite velocity
    theta_slew = inputs['Slew']  # Max slew angle
    Imin = inputs['Imin']        # Minimum moment of inertia
    Imax = inputs['Imax']        # Maximum moment of inertia

    tau_slew = 4*theta_slew*Imax / v['dt_slew']**2
    tau_g = 3 * v['mu'] * np.abs(Imax - Imin) * np.sin(2 * v['theta'] * (np.pi / 180)) / (2 * (v['Re'] + H) ** 3)
    tau_sp = Lsp * Fs * v['As'] * (1 + q) * np.cos(v['thetai']) / v['c']
    tau_m = 2 * v['M'] * v['Rd'] / (v['Re'] + H) ** 3
    tau_a = (1 / 2) * La * v['rhoa'] * Cd * v['A'] * vel ** 2
    tau_dist = np.sqrt(tau_g**2 + tau_sp**2 + tau_m**2 + tau_a**2)
    tau_tot = np.max(np.concatenate((tau_slew[..., np.newaxis], tau_dist[..., np.newaxis]), axis=-1), axis=-1)
    Pacs = tau_tot * (v['omega'] * (2 * np.pi / 60)) + v['nrw'] * v['Phold']

    y = {'Pat': Pacs*pct, 'tau_tot': tau_tot*pct}

    return y


def fire_sat_system(save_dir=None):
    """Fire satellite detection system model from Chaudhuri 2018.

    !!! Note "Some modifications"
        Orbit will save outputs all the time; Power won't because it is part of an FPI loop. Orbit and Power can
        return `np.nan` some of the time (optional, to test robustness of surrogate). Power and Attitude have an
        `alpha` fidelity index that controls accuracy of the returned values. `alpha=0,1,2` corresponds to `8,4,0`
        percent error.

    :param save_dir: where to save model outputs
    :returns: a `System` object for the fire sat MD example system
    """
    return System.load_from_file(Path(__file__).parent / 'fire-sat.yml', root_dir=save_dir)



"""Test variables and variable lists. Including tests for compression, normalization, and distributions."""
import numpy as np
import pytest

from amisc import YamlLoader
from amisc.compression import SVD
from amisc.utils import relative_error
from amisc.variable import Variable, VariableList


def test_load_and_dump_variables(tmp_path):
    """Make sure yaml serialization works for variables."""
    variables = [Variable('a', description='Altitude', units='m', distribution='U(0, 1)'),
                 Variable(u'Φ', description='Base width', units='m', distribution='N(0, 1)'),
                 Variable('p', description='Pressure', units='Pa', domain=(1e6, 2e6),
                          compression={'method': 'svd', 'interpolate_opts': {'kernel': 'gaussian', 'neighbors': 10},
                                       'rank': 10, 'energy_tol': 0.95})
                 ]
    YamlLoader.dump(variables, tmp_path / 'variables.yml')
    variables_load = YamlLoader.load(tmp_path / 'variables.yml')

    for v in variables_load:
        v_old = variables[variables.index(v)]
        assert v_old.get_domain() == v.get_domain()
        assert v_old.description == v.description
        assert v_old.compression == v.compression

    x = Variable()
    y1, y2 = [Variable(), Variable()]
    x_vars = [Variable(distribution='U(0, 1)') for i in range(5)]
    assert x.name == 'x'
    assert y1.name.startswith('X_')
    assert y2.name.startswith('X_')
    assert all([x_vars[i].name.startswith('X_') for i in range(5)])


def test_single_norm():
    norms = ['linear(0.5, 2)', 'log(10)', 'minmax(lb_norm=-1, ub_norm=1)', 'zscore(2, 1)']
    dists = ['U(0.1, 2)', 'N(3, 0.5)']

    for norm_method in norms:
        for dist_method in dists:
            variable = Variable(norm=norm_method, distribution=dist_method)
            x = variable.sample_domain(10)
            xnorm = variable.normalize(x)
            xtilde = variable.normalize(xnorm, denorm=True)
            assert np.allclose(x, xtilde), (f'Failed single normalization test for '
                                            f'distribution: {dist_method}, norm: {norm_method}')


def test_two_norms():
    cases = [['log(offset=1)', 'minmax'], ['linear(0.25, 5)', 'minmax'], ['log', 'linear(-2, 20)'],
             ['zscore(3, 0.5)', 'minmax']]
    dists = ['Uniform(1e-8, 1e-5)', 'Normal(3, 0.5)']

    for dist_method in dists:
        for case in cases:
            var1 = Variable(distribution=dist_method, norm=case)
            var2 = Variable(distribution=dist_method, norm=list(reversed(case)))
            x = var1.sample_domain(10)
            xnorm1 = var1.normalize(x)
            xnorm2 = var2.normalize(x)
            xtilde1 = var1.normalize(xnorm1, denorm=True)
            xtilde2 = var2.normalize(xnorm2, denorm=True)
            assert np.allclose(x, xtilde1), (f'Failed the two-normalization test for distribution: {dist_method}, '
                                             f'norm: {case}')
            assert np.allclose(x, xtilde2), (f'Failed the two-normalization test for distribution: '
                                             f'{dist_method}, norm: {var2.norm}')


def test_many_norms():
    cases = [['linear(0.1)', 'zscore(mu=-1, std=0.25)', 'log(base=10, offset=5)', 'minmax(lb_norm=-1, ub_norm=1)'],
             ['log10(offset=11)', 'minmax', 'linear(-2, 1e-6)']]
    dist = 'U(-10, -5)'
    for case in cases:
        variable = Variable(distribution=dist, norm=case)
        x = variable.sample_domain(50)
        xnorm = variable.normalize(x)
        xtilde = variable.normalize(xnorm, denorm=True)
        assert np.allclose(x, xtilde), f'Failed the many-normalization test for distribution: {dist}, norm: {case}'


def test_nominal_and_domain():
    variable = Variable(distribution='U(1, 10)', norm='log10')
    nominal = variable.get_nominal()
    bds = variable.get_domain()
    assert nominal == (10 + 1)/2
    assert bds == (1, 10)

    variable = Variable(distribution='N(5, 1)', norm='minmax')
    nominal= variable.get_nominal()
    bds = variable.get_domain()
    assert nominal == 5
    assert bds == (2, 8)


def test_compression_1d():
    """Test svd compression for a single-input, single-output, 1d field quantity."""
    A = Variable('A', distribution='U(-5, 5)')
    p = Variable('p')
    def model(inputs, coords):
        amp = inputs['A']
        return {'p': amp[..., np.newaxis] * np.tanh(np.squeeze(coords))}

    # Generate SVD data matrix
    rank = 4
    samples = A.sample(50)
    svd_coords = np.linspace(-2, 2, 200)
    outputs = model({str(A): samples}, svd_coords)
    data_matrix = outputs[p].T  # (dof, num_samples)
    p.compression = SVD(rank=rank, coords=svd_coords, data_matrix=data_matrix)

    # Test compression
    coarse_shape = 25
    coarse_coords = np.linspace(-2, 2, coarse_shape)
    num_test = (5, 20)
    samples = A.sample(num_test)
    outputs = model({str(A): samples}, coarse_coords)
    outputs_reduced = p.compress(outputs, coords=coarse_coords)
    outputs_reconstruct = p.reconstruct(outputs_reduced)

    y = outputs[p]
    yhat = outputs_reconstruct[p]
    assert relative_error(yhat, y) < 0.01

    # Test sampling latent coefficients
    latent_min = np.min(outputs_reduced['latent'], axis=(0, 1))
    latent_max = np.max(outputs_reduced['latent'], axis=(0, 1))
    p.domain = (-1, 1)
    assert len(p.get_domain()) == rank  # inferred from compression rank
    p.domain = [(lmin, lmax) for lmin, lmax in zip(latent_min, latent_max)]
    p.norm = 'linear(0.5)'
    domain = p.get_domain()
    samples = p.sample_domain(num_test)
    nominal = p.sample(num_test)
    lb = np.atleast_1d([d[0] for d in domain])
    ub = np.atleast_1d([d[1] for d in domain])
    assert len(domain) == rank
    assert samples.shape == num_test + (rank,)
    assert nominal.shape == num_test + (rank,)
    assert np.all(np.logical_and(lb < samples, samples < ub))
    assert np.allclose(nominal - (lb+ub)/2, 0)


def test_compression_fields():
    """Test compression for a multi-field quantity."""
    num_grid = 200
    num_samples = 50
    delta = Variable(distribution='U(0, 1)')
    gamma = Variable(distribution='U(1, 2)')
    grid = np.linspace(-1, 1, num_grid)
    delta_samples = delta.sample(num_samples)
    gamma_samples = gamma.sample(num_samples)
    pressure_x = delta_samples[..., np.newaxis] * np.sin(grid)
    pressure_y = gamma_samples[..., np.newaxis] * np.cos(grid)
    pressure_matrix = np.concatenate((pressure_x[..., np.newaxis], pressure_y[..., np.newaxis]),
                                     axis=-1).reshape((num_samples, -1))
    pressure = Variable(compression=SVD(rank=2, data_matrix=pressure_matrix.T, coords=grid,
                                        fields=['pressure_x', 'pressure_y']), name='pressure')

    latent = pressure.compress({'pressure_x': pressure_x, 'pressure_y': pressure_y})
    pressure_reconstruct = pressure.reconstruct(latent)
    assert relative_error(pressure_reconstruct['pressure_x'], pressure_x) < 0.01
    assert relative_error(pressure_reconstruct['pressure_y'], pressure_y) < 0.01

    coarse_grid = 20
    coarse_coords = np.linspace(-1, 1, coarse_grid)
    num_test = (5, 10)
    delta_samples = delta.sample(num_test)
    gamma_samples = gamma.sample(num_test)
    pressure_x = delta_samples[..., np.newaxis] * np.sin(coarse_coords)
    pressure_y = gamma_samples[..., np.newaxis] * np.cos(coarse_coords)

    latent = pressure.compress({'pressure_x': pressure_x, 'pressure_y': pressure_y}, coords=coarse_coords)
    pressure_reconstruct = pressure.reconstruct(latent)

    assert relative_error(pressure_reconstruct['pressure_x'], pressure_x) < 0.01
    assert relative_error(pressure_reconstruct['pressure_y'], pressure_y) < 0.01


@pytest.mark.skip(reason='Not implemented')
def test_compression_nd():
    """Test SVD compression for high-dimensional data."""
    # TODO
    pass


def test_uniform_dist():
    shape = (50,)
    true_lb = [-2, 1.1, 11.3]
    true_ub = [-1, 1.2, 94]
    for i in range(len(true_lb)):
        v = Variable(distribution=f'Uniform({true_lb[i]}, {true_ub[i]})')
        samples = v.sample(shape)
        pdf = v.pdf(samples)
        assert np.all(np.logical_and(true_lb[i] < samples, samples < true_ub[i]))
        assert np.allclose(pdf, 1 / (true_ub[i] - true_lb[i]))
        assert np.allclose(v.pdf(samples + true_ub[i]), 0)

    # Loguniform distribution
    v = Variable(distribution='LogUniform(1e-8, 1e-2)', norm='log10')
    norm_bds = v.normalize(v.get_domain())
    unnorm_bds = v.denormalize(norm_bds)
    samples = v.sample(shape)
    pdf = v.pdf(samples)
    samples_norm = v.normalize(samples)
    assert np.all(np.logical_and(norm_bds[0] < samples_norm, samples_norm < norm_bds[1]))
    assert np.all(np.logical_and(unnorm_bds[0] < samples, samples < unnorm_bds[1]))
    assert np.allclose(pdf, 1 / (samples * (np.log(unnorm_bds[1]) - np.log(unnorm_bds[0]))))


def test_normal_dist():
    shape = (10, 20, 50)
    true_means = [-5.2, 1, 2.7, 14.73, 100]
    true_stds = (np.random.rand(len(true_means)) * 0.1 + 0.05) * np.abs(true_means)
    variables = [Variable(distribution=f'Normal({true_means[i]}, {true_stds[i]})') for i in range(len(true_means))]
    for i, v in enumerate(variables):
        # Plain normal distribution
        samples = v.sample(shape)
        x = np.linspace(*v.get_domain(), 1000)
        pdf = v.pdf(x)
        assert relative_error(np.mean(samples), true_means[i]) < 0.05
        assert relative_error(x[np.argmax(pdf)], true_means[i]) < 0.05

    # Log normal distribution
    v = Variable(distribution='LogNormal(-2, 1, 2.718)', norm='log')
    samples = v.sample(shape)
    x = np.logspace(*v.normalize(v.get_domain()), 1000)
    pdf = v.pdf(x)
    mode = np.exp(-2 - 1**2)
    assert relative_error(np.mean(v.normalize(samples)), -2) < 0.05
    assert relative_error(x[np.argmax(pdf)], mode) < 0.05


def test_relative_and_tolerance_dist():
    shape = (10, 2)
    nominals = [1.1, -3, 1e6, 20.]
    dists = ['rel', 'tol']
    tols = [0.5, 1.1, 10]
    v = Variable()
    for nominal in nominals:
        for dist in dists:
            for tol in tols:
                v.nominal = nominal
                v.distribution = f'{dist}({tol})'
                samples = v.sample(shape)
                tol = tol if dist == 'tol' else (tol/100) * abs(nominal)
                bds = (nominal - tol, nominal + tol)
                assert np.all(np.logical_and(bds[0] < samples, samples < bds[1]))


def test_no_dist():
    shape = (10, 2)
    nominals = [-5.2, 1, 2.7, 14.73, 100]
    v = Variable(domain=(0, 1))
    for i, nominal in enumerate(nominals):
        v.nominal = nominal
        samples = v.sample(shape)
        x = np.linspace(*v.get_domain(), 100)
        pdf = v.pdf(x)
        assert np.allclose(samples, nominal)
        assert np.allclose(pdf, 1)


def test_variable_list(tmp_path):
    # Initialize several variables
    letters = 'abcdefgh'
    means = np.random.rand(len(letters)) * 10 - 5
    std = np.random.rand(len(letters)) * 3
    variables = [Variable(letter, tex=f'${letter}_{i}$', distribution=f'N({means[i]}, {std[i]})', norm='zscore')
                 for i, letter in enumerate(letters)]

    # Init from var, dict, list, or varlist
    l1 = VariableList(variables)
    l2 = VariableList({v: v for v in variables})
    l3 = VariableList(variables[0])
    for v in variables:
        l3.append(v)
    l4 = VariableList(l3)

    for i, v in enumerate(variables):
        assert l1[v] == v
        assert l2.get(v) == v
        assert l3[i] == v
        assert l4[v.name] == v.name

    # Fancy indexing
    indices = [(0, 3, 5), slice(1, 4), ['a', 'b', 'h', 'g', 'f']]
    for index in indices:
        subset1 = l1[index]
        for i, ele in enumerate(subset1):
            assert ele == l2[index][i]
            assert ele == l3[index][i]
            assert ele == l4[index][i]

    # Set/get bad values
    bad_keys = [2.3, np.sum, slice(None), 'a']
    bad_vals = [variables[0], variables[1], variables[2], 5.4]
    for i, key in enumerate(bad_keys):
        try:
            l1[key] = bad_vals[i]
            assert False
        except Exception:
            pass

    bad_keys = ['z', 'y', 'X2', 1.2, np.mean]
    for key in bad_keys:
        try:
            print(l4[key])
            assert False
        except Exception:
            pass

    # Delete and check
    del_idx = (0, -1, 'c')
    l1[-1] = Variable()
    assert l2[-1] != l1[-1]
    del l1[del_idx]
    assert l1[0] == 'b'
    assert l1[-1] == l2[-2]
    assert len(l1) == len(letters) - len(del_idx)

    # Load/dump from yaml
    YamlLoader.dump(l4, tmp_path / 'variable_list.yml')
    variables_load = YamlLoader.load(tmp_path / 'variable_list.yml')

    for v in variables_load:
        v_old = l4[l4.index(v)]
        assert v_old.get_domain() == v.get_domain()
        assert v_old.get_tex(units=True) == v.get_tex(units=True)
